# Copyright 2020-2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Define the graph stream handler."""
from collections import defaultdict
from dataclasses import dataclass, field
from typing import List

from mindinsight.debugger.common.exceptions.exceptions import DebuggerParamValueError, \
    DebuggerNodeNotInGraphError, DebuggerGraphNotExistError
from mindinsight.debugger.common.log import LOGGER as log
from mindinsight.debugger.common.utils import is_scope_type
from mindinsight.debugger.conditionmgr.common.utils import NodeBasicInfo
from mindinsight.debugger.conditionmgr.condition import TargetTypeEnum as CategoryTypeEnum
from mindinsight.debugger.stream_cache.debugger_graph import DebuggerGraph
from mindinsight.debugger.stream_cache.debugger_multigraph import DebuggerMultiGraph
from mindinsight.debugger.stream_handler.base_handler import StreamHandlerBase
from mindinsight.debugger.stream_handler.source_handler import validate_stack_pattern, SourceHandler
from mindinsight.domain.graph.query import StackQuery


class MultiCardGraphHandler(StreamHandlerBase):
    """Multi-card Graph Handler."""

    def __init__(self):
        self._graph_handlers = {}
        self._source_handler = SourceHandler()

    @property
    def graph_handlers(self):
        """The property of whole_graph."""
        return self._graph_handlers

    @property
    def source_handler(self):
        """The property of source handler."""
        return self._source_handler

    def get_graph_handler_by_rank_id(self, rank_id=0):
        """Get handler by rank id."""
        if rank_id in self._graph_handlers:
            return self._graph_handlers.get(rank_id)
        if rank_id == 0:
            # ms side still using device id instead of rank id, so the first rank_id is not 0
            for graph_handler in self._graph_handlers.values():
                return graph_handler
        log.error("There is no rank id %d.", rank_id)
        raise ValueError

    def put(self, value):
        """Put graphs into graph_handlers."""
        for rank_id, graph in value.items():
            if rank_id not in self._graph_handlers:
                self._graph_handlers[rank_id] = GraphHandler()
            self._graph_handlers[rank_id].put(graph)

    def parse_stack_infos(self):
        """Add stack infos into cache."""
        source_handler = self._source_handler
        for graph_stream in self._graph_handlers.values():
            for leaf_node in graph_stream.leaf_nodes.values():
                source_handler.put(leaf_node.stack)
        self._source_handler.sort()

    def get(self, filter_condition=None):
        """Get the graph of specific node for specific device."""
        if not self._graph_handlers:
            log.warning('There is no graph received yet.')
            return {'graph': {}}

        def _get_first_rank_id():
            rank_ids = list(self._graph_handlers.keys())
            rank_ids.sort()
            return rank_ids[0]
        if not filter_condition or not filter_condition.get('rank_id'):
            rank_id = _get_first_rank_id()
        else:
            rank_id = filter_condition.get('rank_id')
        if rank_id in self._graph_handlers:
            return self._graph_handlers.get(rank_id).get(filter_condition)
        log.error("There is no rank id %d.", rank_id)
        raise ValueError

    def has_graph(self):
        """check if has graph"""
        for graph_handler in self._graph_handlers.values():
            if graph_handler.graph_names:
                return True
        return False

    def register_graph_handler(self, rank_id, graph_handler):
        """Register graph handler."""
        self._graph_handlers[rank_id] = graph_handler

    def clean(self):
        """Clean cache."""
        self.__init__()

    def validate_rank_id(self, rank_id):
        """Judge if the rank_id is valid"""
        return rank_id in self._graph_handlers


@dataclass
class RootGraph:
    """The RootGraph data class."""
    graph_name: str = ''
    sub_graph_names: List[str] = field(default_factory=list)


class GraphHandler(StreamHandlerBase):
    """Metadata Handler."""

    def __init__(self):
        # the key is root_graph_id, the value is RootGraph
        self._root_graphs = defaultdict(RootGraph)
        # dict of <graph_name, DebuggerGraph object>
        self._graph = {}
        self._searched_node_list = {}
        # dict of <node full name, graph_name>
        self.graph_node_map = {}
        # dict of <node ui name, Node object> for all graphs
        self._all_leaf_nodes = {}
        # the whole graph
        self._whole_graph = None

    @property
    def whole_graph(self):
        """The property of whole_graph."""
        return self._whole_graph

    @property
    def root_graphs(self):
        """The property of root graphs."""
        return self._root_graphs

    @property
    def graph_names(self):
        """The property of graph names."""
        return list(self._graph)

    @property
    def debugger_graph_obj(self):
        """The property of graph object."""
        return self._graph

    @property
    def leaf_nodes(self):
        """The property of leaf nodes among all graphs."""
        return self._all_leaf_nodes

    def put(self, value):
        """
        Put value into graph cache. Called by grpc server.

        Args:
            value (dict): The Graph proto message. Each item is format like (<graph_name>, GraphProto).
        """
        log.info("Put graph into cache.")
        sorted_value_list = self._sort_graph(value)
        for graph_name, graph_value in sorted_value_list:
            # build sub graph
            graph = DebuggerGraph()
            graph.build_graph(graph_value)
            self._graph[graph_name] = graph
            leaf_nodes = graph.leaf_nodes
            self._all_leaf_nodes.update(leaf_nodes)
            for _, node in leaf_nodes.items():
                self.graph_node_map[node.full_name] = graph_name

        self._parse_root_graphs()
        # build whole graph
        graph = DebuggerMultiGraph()
        graph.add_graph(self._graph)
        self._whole_graph = graph

    def _parse_root_graphs(self):
        """Parse root graph infos."""
        for graph_name, debugger_graph in self._graph.items():
            root_graph_id = debugger_graph.root_graph_id
            root_graph = self._root_graphs[root_graph_id]
            if debugger_graph.is_root_graph:
                log.info("Found root graph: %s", graph_name)
                root_graph.graph_name = graph_name
            else:
                log.info("Found subgraph %s, the root_graph_id is %s", graph_name, root_graph_id)
                root_graph.sub_graph_names.append(graph_name)

    def get(self, filter_condition=None):
        """
        Get the graph of specific node.

        Args:
            filter_condition (dict):

                - name (str): The full debug node name.
                - graph_name (str): The relative graph_name of the node.
                - single_node (bool): If True, return the graph from root
                  to the specific node; else, return the sublayer of the
                  graph. Default: False.

        Returns:
            dict, the metadata.
        """
        try:
            self._graph_exists()
        except DebuggerGraphNotExistError:
            log.warning('The graph is empty. To view a graph, '
                        'please start the training script first.')
            return {'graph': {}}

        graph = {}
        if filter_condition is None:
            filter_condition = {}
            graph = {'graph_names': self.graph_names}

        single_node = filter_condition.get('single_node', False)
        name = filter_condition.get('name')
        graph_name = filter_condition.get('graph_name')
        if single_node is True:
            nodes = self._get_single_node(name, graph_name)
        else:
            nodes = self._list_nodes(name, graph_name)
        graph.update(nodes)

        return {'graph': graph}

    def _get_single_node(self, name, graph_name=None):
        """
        Search node, and return every layer nodes until this node.

        Args:
            graph_name(str): The graph_name.
            name (str): The name of node.

        Returns:
            dict, every layer nodes until this node.
        """
        graph = self._get_graph(graph_name=graph_name)
        searched_graph = graph.search_single_node(name)

        return searched_graph

    def _list_nodes(self, scope, graph_name):
        """
        Get the nodes of every layer in graph.

        Args:
            scope (str): The name of a scope.
            graph_name(str): The graph name.

        Returns:
            TypedDict{'nodes': ['Node_1', ...], 'graph_names': ['graph_name_1', ...]},
            format is {'nodes': [<NodeObject>], 'graph_names': [<str>]}.
                example:
                    {
                      "nodes" : [
                        {
                          "attr" :
                          {
                            "index" : "i: 0\n"
                          },
                          "input" : {},
                          "name" : "input_tensor",
                          "output" :
                          {
                            "Default/TensorAdd-op17" :
                            {
                              "edge_type" : "data",
                              "scope" : "name_scope",
                              "shape" : [1, 16, 128, 128]
                            }
                          },
                          "output_i" : -1,
                          "proxy_input" : {},
                          "proxy_output" : {},
                          "independent_layout" : False,
                          "subnode_count" : 0,
                          "type" : "Data"
                        }
                      ]
                    }
        """
        graph = self._get_graph(graph_name, scope)
        nodes = graph.list_node_by_scope(scope=scope)
        res = {'nodes': nodes}

        return res

    def get_tensor_history(self, node_name, graph_name=None, depth=0):
        """
        Get the tensor history of a specified node.

        Args:
            node_name (str): The debug name of the node.
            graph_name (str): The graph_name. Default: None.
            depth (int): The number of layers the user
                wants to trace. Default is 0.

        Returns:
            dict, basic tensor history, only including tensor name and tensor type and node type.
        """
        graph_name, node_name = self._parse_node_name(node_name, graph_name)
        graph = self._get_graph(graph_name=graph_name, node_name=node_name)
        # validate node type, scope node has no tensor history
        node_type = graph.get_node_type(node_name)
        if is_scope_type(node_type):
            log.error("Scope type node has no tensor history.")
            raise DebuggerParamValueError("Invalid leaf node name.")
        # get tensor history
        tensor_history, cur_outputs_nums = graph.get_tensor_history(node_name, depth)
        # add the tensor type for tensor history
        self._update_tensor_history(tensor_history[0:cur_outputs_nums], 'output', graph_name)
        self._update_tensor_history(tensor_history[cur_outputs_nums:], 'input', graph_name)
        log.debug("Get %d tensors in tensor history for node <%s>.", len(tensor_history), node_name)
        return {'tensor_history': tensor_history}

    @staticmethod
    def _update_tensor_history(tensor_history, tensor_type, graph_name):
        """
        Add tensor source type for tensor history.

        Args:
            tensor_history (list[dict]): Tensor history from Graph stream. Each element has two
                keys: `node_type` and `name`. `node_type` refers to the type of the node which
                the tensor come from. `name` refers to the tensor name.
            tensor_type (str): The source type of the tensor. `input` or `output`.
            graph_name (str): The graph name.
        """
        for single_tensor_info in tensor_history:
            single_tensor_info['type'] = tensor_type
            single_tensor_info['graph_name'] = graph_name

    def search_nodes(self, pattern):
        """
        Search nodes by given pattern.

        Args:
            pattern (dict): Filter condition.

                - name (str): The name pattern.
                - graph_name (str): The graph name.
                - node_category (str): The node_category.
                - condition (dict): The additional filter condition.
                - stack_pattern (str): The pattern of stack info.

        Returns:
            dict, the searched node.
        """
        graph_name = pattern.pop('graph_name', None)
        search_nodes = self.search_in_graph(pattern, graph_name)
        # construct to search tree
        graph = self._get_graph(graph_name=graph_name)
        format_nodes = graph.get_nodes(search_nodes)
        return {'nodes': format_nodes}

    def search_in_graph(self, pattern, graph_name=None):
        """
        Search nodes by given pattern.

        Args:
            pattern (dict): Filter condition.

                - name (str): The name pattern.
                - node_category (str): The node_category. Default: None.
                - condition (dict): The additional filter condition.
                - stack_pattern (str): The pattern of stack info.
            graph_name (str): The graph name. Default: None.

        Returns:
            list, the searched node list.
        """
        temp_node_list = []
        node_category = pattern.get('node_category')
        graph = self._get_graph(graph_name=graph_name)
        stack_pattern = pattern.get('stack_pattern')
        validate_stack_pattern(stack_pattern)
        # filter nodes by name
        if pattern.get('name'):
            if node_category or stack_pattern:
                # get leaf nodes for forward filter
                temp_node_list = graph.search_leaf_nodes_by_pattern(pattern.get('name'))
            else:
                # optimize search nodes
                temp_node_list = graph.search_nodes_by_pattern(pattern.get('name'))
            if not temp_node_list:
                log.debug("No node named %s", pattern.get('name'))
                return []
        # filter nodes by category
        if node_category:
            node_category = self._get_inner_node_category(node_category)
            condition = pattern['condition'].copy() if pattern.get('condition') else {}
            condition['search_range'] = temp_node_list
            temp_node_list = graph.search_nodes_by_category(node_category, condition=condition)
            if not temp_node_list:
                log.debug("No node with category %s", node_category)
                return []
        # filter nodes by stack_pattern
        if stack_pattern:
            if not temp_node_list:
                temp_node_list = graph.leaf_nodes.values()
            log.info("Begin to filter nodes by `stack_pattern`.")
            query = StackQuery(temp_node_list)
            temp_node_list = query.filter(stack_pattern).all()
        return temp_node_list

    @staticmethod
    def _get_inner_node_category(node_category):
        """
        Get inner node category.

        Args:
            node_category (str): The node category supported in
                mindinsight.conditionmgr.condition.TargetTypeEnum.

        Returns:
            CategoryTypeEnum, the translated value.
        """
        try:
            res = CategoryTypeEnum(node_category)
        except ValueError as err:
            log.error("Invalid node category. %s", err)
            raise DebuggerParamValueError("Invalid node_category.")
        return res

    def get_graph_id_by_name(self, node_name):
        """
        Get graph id by full name.

        Args:
            node_name (str): The name of the node.

        Returns:
            str, the graph name of the node.

        Raises:
            DebuggerNodeNotInGraphError: If can not find the node in all graphs.
        """
        if node_name:
            for graph_name, sub_graph in self._graph.items():
                if sub_graph.exist_node(name=node_name):
                    return graph_name
        log.error('Failed to find node %s in graph. Please make sure the graph has been sent and '
                  'the node name is correct, and try again.', node_name)
        raise DebuggerGraphNotExistError

    def get_graph_id_by_full_name(self, node_name):
        """
        Get graph id by full name.

        Args:
            node_name (str): The full name of the node.

        Returns:
            str, the graph name of the node.

        Raises:
            DebuggerNodeNotInGraphError: If can not find the node in all graphs.
        """
        graph_id = self.graph_node_map.get(node_name) if node_name else None
        if not graph_id:
            log.warning("Failed to get graph id by full name: %s", node_name)
        return graph_id

    def get_node_type(self, node_name, graph_name=None):
        """
        Get the type of the specified node.

        Args:
            node_name (str): The debug name of the node.
            graph_name (str): The relative graph_name of the node. Default: None.

        Returns:
            A string of the node type, name_scope or leaf.
        """
        if graph_name:
            graph = self._get_graph(node_name=node_name, graph_name=graph_name)
        else:
            graph = self._whole_graph
        node_type = graph.get_node_type(node_name)

        return node_type

    def get_full_name(self, node_name, graph_name=None):
        """Get full name according to ui node name."""
        full_name = ''
        if node_name:
            graph = self._get_graph(node_name=node_name, graph_name=graph_name)
            full_name = graph.get_full_name_by_node_name(node_name)

        return full_name

    def get_node_basic_info(self, node_name, graph_name):
        """Get node basic info with graph scope."""
        graph_name, node_name = self._parse_node_name(node_name=node_name, graph_name=graph_name)
        graph = self._get_graph(graph_name, node_name)
        full_name = graph.get_full_name_by_node_name(node_name)
        node_type = graph.get_node_type(node_name)
        return self.construct_node_basic_info(full_name, graph_name, node_name, node_type)

    def get_tensor_graph(self, tensor_name, graph_name):
        """
        Get tensor graph according to node name.

        Args:
            tensor_name (str): Tensor name from UI, format is "node_name:slot".
            graph_name (str): The relative graph_name of the node. Default: None.

        Returns:
            dict, relative node.
        """
        node_name, _ = tensor_name.rsplit(':', 1)
        graph = self._get_graph(graph_name=graph_name, node_name=node_name)
        tensor_graph = graph.get_tensor_graph(node_name)
        return {'graph': tensor_graph}

    @staticmethod
    def construct_node_basic_info(full_name, graph_name, node_name, node_type):
        """Construct node basic info."""
        node_name_with_graph_scope = '/'.join([graph_name, node_name]) if node_name else graph_name
        return NodeBasicInfo(name=node_name_with_graph_scope, full_name=full_name, type=node_type)

    def get_node_basic_info_by_scope(self, scope_name, graph_name):
        """
        Get node by a given scope name.

        Args:
            scope_name (str): The name of scope.
            graph_name (str): The relative graph_name of the watched node. Default: None.

        Returns:
            list[NodeBasicInfo], a list of node.
        """
        graph_name, node_name = self._parse_node_name(scope_name, graph_name)
        graph = self._get_graph(graph_name)
        # to make sure fully match the scope name
        node_name = node_name + '/' if node_name and not node_name.endswith('/') else node_name
        nodes = graph.search_leaf_nodes_by_pattern(node_name, True)
        res = [self.construct_node_basic_info(full_name=node.full_name,
                                              graph_name=graph_name,
                                              node_name=node.name,
                                              node_type=node.type) for node in nodes]
        return res

    def get_node_name_by_full_name(self, full_name, graph_name):
        """Get UI node name by full name and graph name."""
        if graph_name and full_name:
            graph = self._get_graph(graph_name)
            node_name = graph.get_node_name_by_full_name(full_name)
        else:
            node_name = ''
            log.debug("Get empty full name.")
        return node_name

    def _graph_exists(self):
        """
        Check if the graph has been loaded in the debugger cache.

        Raises:
            DebuggerGraphNotExistError: If the graph does not exist.
        """
        if not self._graph:
            log.error('The graph does not exist. Please start the '
                      'training script and try again.')
            raise DebuggerGraphNotExistError

    def _get_graph(self, graph_name=None, node_name=None):
        """
        Get the graph object according to graph name and node name.

        Args:
            graph_name (str): The graph name.
            node_name (str): The node name.

        Returns:
            DebuggerGraph, the graph object.

        Raises:
            DebuggerGraphNotExistError: If the graph does not exist.
        """
        graph = self._graph.get(graph_name) if graph_name else self._whole_graph
        # get graph according to graph name and check the node
        if graph and (not node_name or graph.exist_node(name=node_name)):
            return graph
        log.error('The graph %s does not exist node %s.', graph_name, node_name)
        raise DebuggerGraphNotExistError

    def _has_graph_scope(self, graph_name):
        """Check if query with graph_scope."""
        return bool(graph_name is None and len(self._graph) > 1)

    def validate_graph_name(self, graph_name):
        """Validate graph_name."""
        if graph_name and self._graph.get(graph_name) is None:
            log.error("No graph named %s in debugger cache.", graph_name)
            raise DebuggerGraphNotExistError
        if not graph_name and len(self._graph) == 1:
            graph_name = self.graph_names[0]
        return graph_name

    def _add_graph_scope_for_nodes(self, nodes, graph_name):
        """
        Add graph scope for nodes.

        Args:
            nodes (list[Node]): List of nodes object.
            graph_name (str): The graph name.
        """

        def _get_updated_node_info(cur_node, node_type):
            """Add graph scope in key."""
            old_node = cur_node.get(node_type)
            if not old_node:
                return
            new_values = {}
            for old_name, node_info in old_node.items():
                new_name = '/'.join([graph_name, old_name]) if old_name else graph_name
                new_values[new_name] = node_info
            cur_node[node_type] = new_values

        for node in nodes:
            node['name'] = '/'.join([graph_name, node['name']]) if node['name'] else graph_name
            _get_updated_node_info(node, 'input')
            _get_updated_node_info(node, 'output')
            if node.get('nodes'):
                self._add_graph_scope_for_nodes(node.get('nodes'), graph_name)

    def _parse_node_name(self, node_name, graph_name):
        """
        Parse node_name according to graph_name.

        Args:
            node_name (str): The ui node name.
            graph_name (str): The graph name.

        Returns:
            str, parsed graph name.
            str, parsed node name.
        """
        node_name = '' if node_name is None else node_name
        if self._has_graph_scope(graph_name):
            names = node_name.split("/", 1)
            graph_name = names[0]
            node_name = names[1] if len(names) == 2 else ''
        if graph_name is None and len(self._graph) == 1:
            graph_name = self.graph_names[0]
        return graph_name, node_name

    def validate_node_name(self, node_name, graph_name):
        """
        Validate the graph exist the specified node.

        Args:
            node_name (str): The ui node name.
            graph_name (str): The graph name.

        Raises:
            DebuggerNodeNotInGraphError: If can not find the node in all graphs.
        """
        graph = self._get_graph(graph_name=graph_name)
        if not graph.exist_node(name=node_name):
            log.error("graph %s doesn't find node: %s.", graph_name, node_name)
            raise DebuggerNodeNotInGraphError(node_name)

    @staticmethod
    def _sort_graph(graphs):
        """
        Sort graph by graph_name.

        Args:
            graphs(dict): <graph_name, GraphProto object>.
        """
        if len(graphs) == 1:
            return graphs.items()
        sorted_graphs = sorted(graphs.items(), key=lambda x: get_graph_number(x[0]))
        return sorted_graphs

    def get_root_graph_id(self, graph_name, node_name=None):
        """Get the root_graph_id of a graph."""
        graph_name, node_name = self._parse_node_name(node_name=node_name, graph_name=graph_name)
        graph = self._get_graph(graph_name=graph_name)
        return graph.root_graph_id


def get_graph_number(graph_name):
    """Get the number of graph."""
    number = graph_name.split("_")[-1]
    return int(number)



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Define the GraphHistory stream handler."""
from mindinsight.debugger.common.exceptions.exceptions import DebuggerParamValueError
from mindinsight.debugger.common.log import LOGGER as log
from mindinsight.debugger.common.utils import enter_tag
from mindinsight.debugger.stream_cache.graph_history import GraphHistory
from mindinsight.debugger.stream_handler.base_handler import StreamHandlerBase


class GraphHistoryHandler(StreamHandlerBase):
    """GraphHistory Handler."""

    def __init__(self):
        # the key is rank id, the value is GraphHistory
        self._history = {}
        # the key is rank id, the value is like Dict[<graph_id>, List[<step_id>]]
        self._dumped_step = {}

    def put(self, value):
        """
        Put graph history into cache.

        Args:
            value (dict): The key is the rank id, the value is like {<graph_id>: list[<step_id>]}.
        """
        for rank_id, graph_history in value.items():
            self._history[rank_id] = GraphHistory(rank_id, graph_history)

    def put_dumped_step(self, dumped_step):
        """
        Put dumped step info.

        Args:
            dumped_step (dict): The key is rank id, the
                value is like Dict[<graph_id>, List[<step_id>]].
        """
        self._check_dumped_step(dumped_step)
        self._dumped_step = dumped_step

    @enter_tag
    def _check_dumped_step(self, dumped_step):
        """Check if all dumped step in is graph history."""
        def _matched(dumped_steps, run_history):
            diff_set = dumped_steps - run_history
            if not diff_set:
                return True
            if len(diff_set) != 1:
                return False
            # check if it is a data sink exception case
            max_step_id = max(dumped_steps)
            if max_step_id in diff_set and (max_step_id - 1) in run_history:
                msg = f"The dumped iteration {max_step_id} is not expected. It may be normal if " \
                          f"the data is dumped in dataset sink mode and there is only AssignAdd operator " \
                          f"in that directory. Otherwise, the dump structure is not correct."
                log.warning(msg)
                dumped_steps.remove(max_step_id)
                return True
            return False

        for rank_id, value in dumped_step.items():
            graph_history = self.get(rank_id)
            is_matched = True
            for root_graph_id, dumped_steps_per_graph in value.items():
                whole_run_ids = graph_history.get_run_ids(root_graph_id)
                if not whole_run_ids:
                    # in case there is no graph history file
                    log.warning("Graph History file of %s in rank_%s is empty.", root_graph_id, rank_id)
                    continue
                is_matched = is_matched and _matched(dumped_steps_per_graph, whole_run_ids)
                if not is_matched:
                    log.warning("Dumped step value of graph %s in rank_%s mismatch the graph history. "
                                "Clear the history graph.", root_graph_id, rank_id)
                    graph_history.clear()

    def get(self, filter_condition=None):
        """
        Get GraphHistory of specified rank id.

        Args:
            filter_condition (int): The rank id.

        Returns:
            GraphHistory, the GraphHistory.
        """
        rank_id = filter_condition
        if rank_id not in self._history:
            log.error("Invalid rank id. Failed to find graph history with rank id %s", rank_id)
            raise DebuggerParamValueError("Invalid rank_id.")
        return self._history.get(rank_id)

    def get_dumped_step(self, rank_id):
        """
        Get dumped step info of specified device.

        Args:
            rank_id (int): The rank id.

        Returns:
            dict, the key is graph id, the value is the list of dumped steps.
        """
        if rank_id not in self._dumped_step:
            log.error("Invalid rank id. Failed to find dumped steps with rank id %s", rank_id)
            raise DebuggerParamValueError("Invalid rank_id.")
        return self._dumped_step.get(rank_id)

    def get_total_count(self):
        """
        Get total count per device.

        Returns:
            dict, the key is rank id, the value is the total count of each device.
        """
        total_count = {}
        for rank_id, graph_history in self._history.items():
            total_count_per_rank = graph_history.get_total_count()
            if not total_count_per_rank:
                log.warning("Failed to get total graph run count from graph history of rank_%s. "
                            "Try to get it based on dumped iterations.", rank_id)
                total_count_per_rank = self._get_max_dumped_step_id(rank_id)
            total_count[rank_id] = total_count_per_rank
        return total_count

    def _get_max_dumped_step_id(self, rank_id):
        """Get the maximum step id of specified device."""
        max_step_id = -1
        for dumped_step_per_graph in self._dumped_step.get(rank_id, {}).values():
            if not dumped_step_per_graph:
                continue
            max_step_id = max(max_step_id, max(dumped_step_per_graph))
        return max_step_id + 1



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""This module is aimed to deal with controlling commands."""
import enum

from mindinsight.debugger.common.exceptions.exceptions import DebuggerContinueError, DebuggerParamValueError, \
    DebuggerPauseError, DebuggerRecheckError, DebuggerStepNumError
from mindinsight.debugger.common.log import LOGGER as log
from mindinsight.debugger.common.utils import Streams, get_ack_reply, ServerStatus, RunLevel, is_scope_type, \
    DebuggerServerMode
from mindinsight.debugger.proto.debug_grpc_pb2 import RunCMD
from mindinsight.utils.exceptions import MindInsightException


@enum.unique
class ControlTypeEnum(enum.Enum):
    """Control Type."""
    CONTINUE = 'continue'  # continue to run training
    PAUSE = 'pause'  # suspend training
    TERMINATE = 'terminate'  # terminate training
    RESET = 'reset'  # reset the step_id in offline debugger


class TrainingControlOperator:
    """Control training operator."""
    # max step number should be less than int32
    _MAX_STEP_NUM = 2 ** 31 - 1

    def __init__(self, cache_store):
        self._cache_store = cache_store
        self._watchpoint_stream = cache_store.get_stream_handler(Streams.WATCHPOINT)
        self._multi_card_graph_stream = cache_store.get_stream_handler(Streams.GRAPH)
        self._metadata_stream = cache_store.get_stream_handler(Streams.METADATA)

    @staticmethod
    def validate_mode(mode):
        """Validate mode."""
        enum_members = [item.value for item in ControlTypeEnum]
        if mode not in enum_members:
            log.error("Invalid control mode %s", mode)
            raise DebuggerParamValueError("Invalid control mode.")

    def control(self, mode, params):
        """
        Control the training process.

        Args:
            mode (str): Acceptable control command, including `continue`,
                    `pause` and `terminate`.
            params (dict): The control params.

                - level (str): The control granularity, `node` level or `step` level.
                  Default: `step`.
                - steps (int): Specify the steps that training should run.
                  Used when `level` is `step`.
                - name (str): Specify the name of the node. Used when `level` is `node`.
                - graph_name (str): The graph name.

        Returns:
            dict, the response.
        """
        if mode == ControlTypeEnum.CONTINUE.value:
            reply = self.continue_training(params)
        elif mode == ControlTypeEnum.RESET.value:
            step_id = params['steps']
            reply = self.reset_training_step(step_id)
        else:
            mode_mapping = {
                ControlTypeEnum.PAUSE.value: self.pause_training,
                ControlTypeEnum.TERMINATE.value: self.terminate_training
            }
            reply = mode_mapping.get(mode)()
        return reply

    def continue_training(self, params):
        """
        Send RunCMD to MindSpore.

        Args:
            params (dict): The control params.

        Returns:
            dict, metadata info.
        """
        metadata_stream = self._metadata_stream
        if metadata_stream.state != ServerStatus.WAITING.value:
            log.error("MindSpore is not ready to run. Current state is: %s", metadata_stream.state)
            raise DebuggerContinueError(
                "MindSpore is not ready to run or is running currently."
            )
        metadata_stream.state = ServerStatus.SENDING.value
        try:
            self._validate_continue_params(params)
            event = self._construct_run_event(params)
            # whether need to send recheck before continue, especially for initialization watchpoint
            recheck_flag = bool(self._metadata_stream.step == 0 and self._watchpoint_stream.is_recheckable())
            self._send_watchpoints()
            if recheck_flag:
                self._cache_store.put_command(self._construct_run_event({'level': 'recheck'}))
                log.info("Send recheck command for initialization watchpoints before continue command.")
            self._cache_store.put_command(event)
        except MindInsightException as err:
            log.error("Failed to send run event.")
            log.exception(err)
            metadata_stream.state = ServerStatus.WAITING.value
            raise DebuggerContinueError("Failed to send run command.")
        else:
            metadata_stream.enable_recheck = False
            log.debug("Send the RunCMD to command queue.")
        return metadata_stream.get(['state', 'enable_recheck'])

    def _validate_continue_params(self, params):
        """
        Validate continue params.

        Args:
            params (dict): The control params.

                - level (str): The control granularity, `node`, `step` or `recheck` level.
                    Default: `step`.
                - steps (int): Specify the steps that training should run.
                    Used when `level` is `step`.
                - name (str): Specify the name of the node. Used when `level` is `node`.
                - graph_name (str): The graph name.

        Raises:
            DebuggerParamValueError: Params are invalid.
            DebuggerStepNumError: Step number are invalid.
        """
        # validate level
        level = params.get('level', 'step')
        if level not in [RunLevel.NODE.value, RunLevel.STEP.value, RunLevel.RECHECK.value]:
            log.error("Invalid Value. `level` should be `step`, `node` or `recheck`. Got %s", level)
            raise DebuggerParamValueError("level` should be `step`, `node` or `recheck`.")

        # validate steps
        step_num = params.get('steps', 1)
        if not isinstance(step_num, int) or not (step_num == -1 or 0 < step_num <= self._MAX_STEP_NUM):
            log.error("Invalid step value. Step number should be integer and in [1, 2^31 - 1] or -1.")
            raise DebuggerStepNumError

        # validate node name
        if level == RunLevel.NODE.value:
            node_name = params.get('name')
            graph_name = params.get('graph_name')
            rank_id = params.get('rank_id', 0)
            self._validate_continue_node_name(node_name, graph_name, rank_id)

    def _validate_continue_node_name(self, node_name, graph_name, rank_id):
        """Validate if the node is a leaf node."""
        if not node_name:
            return
        node_type = self._multi_card_graph_stream.get_graph_handler_by_rank_id(rank_id).get_node_type(node_name,
                                                                                                      graph_name)
        if is_scope_type(node_type):
            log.error("Scope type node has no tensor history.")
            raise DebuggerParamValueError("Invalid leaf node name.")

    def _construct_run_event(self, params):
        """
        Construct run cmd from input control params.

        Args:
            params (dict): The control params.

                - level (str): The control granularity, `node`, `step` or `recheck` level.
                    Default: `step`.
                - steps (int): Specify the steps that training should run.
                    Used when `level` is `step`.
                - name (str): Specify the name of the node. Used when `level` is `node`.
                - graph_name (str): The graph name.

        Returns:
            EventReply, control event with run command.
        """
        level = params.get('level', 'step')
        # construct run command events
        event = get_ack_reply()
        if level == 'step':
            steps = params.get('steps', 1)
            run_cmd = RunCMD(run_level='step', run_steps=steps)
        elif level == 'node':
            name = params.get('name', '')
            graph_name = params.get('graph_name')
            if name:
                rank_id = params.get('rank_id', 0)
                name = self._multi_card_graph_stream.get_graph_handler_by_rank_id(rank_id).get_full_name(name,
                                                                                                         graph_name)
            run_cmd = RunCMD(run_level='node', node_name=name)
        else:
            run_cmd = RunCMD(run_level='recheck')

        event.run_cmd.CopyFrom(run_cmd)
        log.debug("Construct run event. %s", event)
        return event

    def _send_watchpoints(self):
        """Send watchpoints to client."""
        set_commands = self._watchpoint_stream.get_pending_commands()
        if not set_commands:
            return
        for set_cmd in set_commands:
            event = get_ack_reply()
            event.set_cmd.CopyFrom(set_cmd)
            self._cache_store.put_command(event)
            log.debug("Send SetCMD to MindSpore. %s", event)

    def pause_training(self):
        """
        Pause the training.

        Returns:
            dict, metadata info.
        """
        metadata_stream = self._metadata_stream
        if metadata_stream.state != ServerStatus.RUNNING.value:
            log.error("The MindSpore is not running.")
            raise DebuggerPauseError("The MindSpore is not running.")
        metadata_stream.state = ServerStatus.SENDING.value
        event = get_ack_reply()
        event.run_cmd.CopyFrom(RunCMD(run_level='step', run_steps=0))
        self._cache_store.clean_command()
        self._cache_store.put_command(event)
        metadata_stream.enable_recheck = False
        log.debug("Send the Pause command")
        return metadata_stream.get(['state', 'enable_recheck'])

    def terminate_training(self):
        """
        Terminate the training.

        Returns:
            dict, metadata info.
        """
        metadata_stream = self._metadata_stream
        metadata_stream.state = ServerStatus.SENDING.value
        self._cache_store.clean_data()
        self._cache_store.clean_command()
        event = get_ack_reply()
        event.exit = True
        self._cache_store.put_command(event)
        metadata_stream.enable_recheck = False
        log.debug("Send the ExitCMD.")
        return metadata_stream.get(['state', 'enable_recheck'])

    def recheck(self):
        """
        Recheck all watchpoints.

        Returns:
            dict, metadata info.
        """
        metadata_stream = self._metadata_stream
        # validate backend status is able to recheck watchpoint
        if not metadata_stream.enable_recheck:
            log.error("Recheck is not available.")
            raise DebuggerRecheckError("Recheck is not available.")
        metadata_stream.state = ServerStatus.SENDING.value
        metadata_stream.enable_recheck = False
        # send updated watchpoint and recheck command
        try:
            event = self._construct_run_event({'level': 'recheck'})
            self._send_watchpoints()
            self._cache_store.put_command(event)
        except MindInsightException as err:
            log.error("Failed to send recheck event.")
            log.exception(err)
            metadata_stream.state = ServerStatus.WAITING.value
            metadata_stream.enable_recheck = True
            raise DebuggerContinueError("Failed to send recheck command.")
        else:
            log.debug("Send the recheck to command queue.")
        return metadata_stream.get(['state', 'enable_recheck'])

    def reset_training_step(self, step_id):
        """
        Reset the training step.

        Args:
            step_id (int): The target step_id.

        Returns:
            dict, metadata info.
        """
        metadata_stream = self._metadata_stream
        if metadata_stream.debugger_type == DebuggerServerMode.ONLINE.value:
            log.error("'step_id' can not be changed manually in online debugger.")
            return metadata_stream.get(['state', 'enable_recheck', 'step'])
        if step_id > metadata_stream.max_step_num:
            log.error("Invalid step_id, step_id should be less than %d.", metadata_stream.max_step_num)
            raise DebuggerParamValueError("Invalid step_id.")
        metadata_stream.state = ServerStatus.SENDING.value
        metadata_stream.step = step_id
        self._cache_store.get_stream_handler(Streams.TENSOR).set_step(step_id)
        self._cache_store.clean_data()
        self._cache_store.clean_command()
        metadata_stream.enable_recheck = True
        metadata_stream.state = ServerStatus.WAITING.value
        self._cache_store.get_stream_handler(Streams.WATCHPOINT).set_outdated()
        log.debug("Send the Change_training_step CMD.")
        return metadata_stream.get(['state', 'enable_recheck', 'step'])



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""The ParallelStrategyAnalyser analyser class."""
import os
import stat
import json
import time
import threading
from pathlib import Path
from enum import Enum
from collections import defaultdict, namedtuple
from multiprocessing import Pool
from marshmallow import ValidationError

from mindinsight.profiler.common.log import logger
from mindinsight.profiler.common.exceptions.exceptions import NotFoundParallelStrategyDataException
from mindinsight.profiler.common.exceptions.exceptions import WrongParallelStrategyDataException
from mindinsight.profiler.common.exceptions.exceptions import ParseParallelStrategyDataException
from mindinsight.profiler.common.exceptions.exceptions import UnsupportedParallelTypeException
from mindinsight.profiler.common.exceptions.exceptions import FileNumNotMatchException
from mindinsight.profiler.common.exceptions.exceptions import ProfilerPathErrorException
from mindinsight.utils.exceptions import MindInsightException
from mindinsight.profiler.common.validator.validate_path import validate_and_normalize_path
from mindinsight.profiler.common.util import get_parallel_message

from .graph import GraphManager, Graph, SupportedParallelType


class Status(Enum):
    PENDING = 'pending'
    LOADING = 'loading'
    FINISH = 'finish'


FileData = namedtuple('FileData', ['graph_proto', 'rank_id'])
STRATEGY_FILE_PATTERN = "parallel_strategy*.json"


class ParallelStrategyCache:
    """Cache the ParallelStrategy Data"""
    _paraller_data = defaultdict(dict)
    _lock = threading.Lock()
    _graph_file = "/graph_{}"

    @classmethod
    def set_cache(cls, key, data, profiler_dir, save=False):
        """cache parallel data"""
        try:
            with cls._lock:
                cls._paraller_data[profiler_dir].update({key: data})
            if save:
                sub_graph_path = profiler_dir + cls._graph_file.format(key)
                flags = os.O_WRONLY | os.O_CREAT | os.O_EXCL
                modes = stat.S_IWGRP | stat.S_IWOTH | stat.S_IWUSR
                with os.fdopen(os.open(sub_graph_path, flags, modes), "w") as fw:
                    json.dump(data, fw)
        except FileExistsError as exc:
            logger.warning("Cache parallel strategy data failed, detail: %s.", str(exc))
        except Exception as exc:
            logger.exception("Cache parallel strategy data failed, detail: %s.", str(exc))

    @classmethod
    def get_cache(cls, key, profiler_dir):
        """get cached parallel data"""
        try:
            res_data = cls._paraller_data.get(profiler_dir, {}).get(key, {})
            sub_graph_path = profiler_dir + cls._graph_file.format(key)
            if not res_data and os.path.exists(sub_graph_path):
                logger.warning("get data from l2 cache")
                with open(sub_graph_path, 'r') as fr:
                    res_data = json.load(fr)
                thread = threading.Thread(target=cls.set_cache, args=(key, res_data, profiler_dir))
                thread.start()
            elif res_data:
                logger.warning("get data from l1 cache")
                if not os.path.exists(sub_graph_path):
                    thread = threading.Thread(target=cls.set_cache, args=(key, res_data, profiler_dir, True))
                    thread.start()
            else:
                logger.warning("data not cached")
            return res_data

        except Exception as exc:
            logger.exception("Get cached parallel strategy data failed, detail: %s.", str(exc))
            return {}


class ParallelStrategyAnalyser:
    """The analyser for parallel strategy."""

    _instance = None
    _lock = threading.Lock()
    _should_init = None

    def __new__(cls, *args):
        """Create a new object."""
        if cls._instance is None:
            with cls._lock:
                if cls._instance is None:
                    cls._instance = super().__new__(cls)
                    setattr(cls._instance, '_should_init', True)
                return cls._instance

        with cls._lock:
            if cls._is_create_new_instance(args[0]):
                cls._instance = super().__new__(cls)
                setattr(cls._instance, '_should_init', True)
                return cls._instance

        setattr(cls._instance, '_should_init', False)
        return cls._instance

    def __init__(self, profiler_dir):
        if not self._should_init:
            return
        self._status = {}
        self._data = {}
        self._profiling_dir = self._normalize_profiling_dir(profiler_dir)
        self._profiler_dir_mtime = self._get_strategy_file_mtime(self._profiling_dir)
        self._exception = None

        parallel_message = get_parallel_message(self._profiling_dir)
        self.parallel_type = parallel_message.get('parallel_mode')
        self.stage_num = parallel_message.get('stage_num')
        self.rank_size = parallel_message.get('rank_size')
        self.group_num = self.rank_size // self.stage_num
        self.stage_devices = {}
        for stage_id in range(self.stage_num):
            self._status.update({stage_id: Status.PENDING.value})
            start = stage_id * self.group_num
            end = start + self.group_num
            self.stage_devices[stage_id] = list(range(start, end))
        self._status.update({'metadata': Status.FINISH.value})
        self.set_data('metadata', {'parallel_type': self.parallel_type, 'stage_devices': self.stage_devices})

        path = Path(self._profiling_dir)
        self.files = sorted(path.glob(STRATEGY_FILE_PATTERN),
                            key=lambda filepath: int(filepath.name.split('.')[0].split('_')[-1]))
        if not self.files:
            logger.error("Can not find any data in path %s", path.name)
            raise NotFoundParallelStrategyDataException()
        if len(self.files) != self.rank_size:
            logger.error("The num of parallel data is not %s , in path %s", str(self.rank_size), path.name)
            raise FileNumNotMatchException()

        if self.parallel_type not in SupportedParallelType.list_members():
            raise UnsupportedParallelTypeException(f"Only support {SupportedParallelType.list_members()} "
                                                   f"parallel type, but current parallel type is {self.parallel_type}.")

    @classmethod
    def _is_create_new_instance(cls, profiler_dir):
        """Determine whether to create a new instance object."""
        # Note: This could be due to multithreading. As a result, instance has not been initialized successfully
        if getattr(cls._instance, '_profiling_dir', None) is None:
            return False
        if getattr(cls._instance, '_profiling_dir') != profiler_dir:
            return True
        if getattr(cls._instance, '_profiler_dir_mtime') != cls._get_strategy_file_mtime(profiler_dir):
            return True
        return False

    @staticmethod
    def _get_strategy_file_mtime(profiling_dir):
        """Get strategy file mtime from profiling dir."""
        path = Path(profiling_dir)
        strategy_files = sorted(path.glob(STRATEGY_FILE_PATTERN), key=lambda file: os.path.getmtime(file))
        if strategy_files:
            file_path = os.path.join(profiling_dir, strategy_files[-1])
            return os.path.getmtime(file_path)
        return -1

    @staticmethod
    def _build_graph(file_data):
        """This function is used for multiprocessing to build graph."""
        graph = Graph()
        try:
            graph.build_graph(graph_proto=file_data.graph_proto, rank_id=file_data.rank_id)
        except Exception as exc:
            return graph, file_data.rank_id, exc
        return graph, file_data.rank_id, None

    @staticmethod
    def _load_file(file: Path):
        """Load data from one file."""
        try:
            with file.open('r') as fp:
                parallel = json.load(fp)
                return parallel, None
        except OSError as exc:
            return None, exc
        except json.JSONDecodeError:
            return None, WrongParallelStrategyDataException(f"The file is not a valid json file: {file.name}.")

    @staticmethod
    def _normalize_profiling_dir(profiling_dir):
        """Normalize the profiling dir."""
        try:
            return validate_and_normalize_path(profiling_dir, 'profiler')
        except ValidationError:
            raise ProfilerPathErrorException('The profiling dir is invalid.')

    def set_data(self, stage_id, data):
        self._data[stage_id] = data
        thread = threading.Thread(target=ParallelStrategyCache.set_cache,
                                  args=(stage_id, data, self._profiling_dir, True))
        thread.start()

    def get_data(self, stage_id):
        """Get the one of stages data."""

        if self._exception is not None:
            raise self._exception

        response = {'status': self._status.get(stage_id), 'data': {}}
        if self._status.get(stage_id) == Status.FINISH.value:
            data = self._data.get(stage_id, {})
            if not data:
                raise WrongParallelStrategyDataException("Can not find any data or load data failed.")
            response.update({'data': data})
        return response

    def load_by_stage_id(self, stage_id):
        """Load the parallel data by stage_id"""
        if self._status.get(stage_id) in (Status.LOADING.value, Status.FINISH.value):
            return
        thread = threading.Thread(target=self._load_data_with_catch_exception,
                                  args=(stage_id, self.stage_devices.get(stage_id)))
        thread.start()

    def _load_data_with_catch_exception(self, stage_id, stage_devices):
        """Wrap load data with try"""
        start = time.time()
        self._status[stage_id] = Status.LOADING.value
        logger.info('Start to load data, status: %s.', self._status.get(stage_id))
        try:
            # Consider processing data from up to 8 cards at a time
            with Pool(processes=8) as pool:
                self._load_data(pool, stage_id, stage_devices)
        except MindInsightException as exc:
            self._exception = exc
        except Exception as exc:
            self._exception = ParseParallelStrategyDataException(str(exc))
            logger.exception("Load parallel strategy data failed, detail: %s.", str(exc))
        finally:
            self._status[stage_id] = Status.FINISH.value
            logger.info("Load data from %s use time: %s", self._profiling_dir, (time.time() - start))

    def _load_data(self, pool, stage_id, stage_devices):
        """Load data in thread."""
        start = stage_id * self.group_num
        end = start + self.group_num
        files = self.files[start: end]

        file_data_dict = {}
        for parallel, exc in pool.imap(self._load_file, files):
            if exc:
                raise exc

            if 'config' not in parallel or not isinstance(parallel.get('config'), dict):
                raise WrongParallelStrategyDataException("Can not find 'config' key in the given data.")

            rank_id = parallel['config']['rankId']
            if rank_id in file_data_dict:
                raise WrongParallelStrategyDataException(f"The file with the same rank id was found. "
                                                         f"There should be no file names with the same rank id.")

            file_data_dict[rank_id] = FileData(graph_proto=parallel['graph'], rank_id=rank_id)

        graph_manager = GraphManager(self.parallel_type, stage_id, stage_devices)
        for graph, rank_id, exc in pool.imap(self._build_graph, file_data_dict.values()):
            if exc:
                raise exc
            graph_manager.add_graph(graph, rank_id)

        graph_manager.merge_graph()
        self.set_data(stage_id, graph_manager.to_dict())
        self._status[stage_id] = Status.FINISH.value



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""The proposer factory."""
import threading

import mindinsight.profiler.proposer.allproposers as proposer_module
from mindinsight.profiler.common.log import logger


class ProposerFactory:
    """The Proposer factory is used to create Proposer special instance."""
    _lock = threading.Lock()
    _instance = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            with cls._lock:
                if cls._instance is None:
                    cls._instance = super().__new__(cls, *args, **kwargs)
        return cls._instance

    @classmethod
    def instance(cls):
        """The factory instance."""
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def get_proposer(self, proposer_type, *args):
        """
        Get the specified proposer according to the proposer type.

        Args:
            proposer_type (str): The proposer type.
            args (list): The parameters required for the specific proposer class.

        Returns:
            Proposer, the specified proposer instance.

        Examples:
            >>> proposer_type = 'step_trace'
            >>> proposer = ProposerFactory.instance().get_proposer(proposer_type, self.profiling_dir, self.device_id)

        """
        logger.debug("The 'proposer_type' is %s,The 'args' is %s", proposer_type, str(args))
        proposer_instance = None
        sub_name = proposer_type.split('_')
        proposer_class_name = ''.join([name.capitalize() for name in sub_name])
        proposer_class_name += 'Proposer'

        if hasattr(proposer_module, proposer_class_name):
            proposer_instance = getattr(proposer_module, proposer_class_name)(*args)
        else:
            logger.warning("The proposer class %s does not exist.", proposer_class_name)
        return proposer_instance



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Cache item updater."""
import os

from mindinsight.datavisual.data_transform.data_manager import BaseCacheItemUpdater, CachedTrainJob
from mindinsight.lineagemgr.common.log import logger
from mindinsight.lineagemgr.common.exceptions.exceptions import LineageFileNotFoundError
from mindinsight.lineagemgr.common.validator.validate import validate_train_id, validate_added_info
from mindinsight.lineagemgr.lineage_parser import LineageParser, LINEAGE
from mindinsight.utils.exceptions import ParamValueError


def update_lineage_object(data_manager, train_id, added_info: dict):
    """Update lineage objects about tag and remark."""
    validate_train_id(train_id)
    validate_added_info(added_info)
    cache_item = data_manager.get_brief_train_job(train_id)
    lineage_item = cache_item.get(key=LINEAGE, raise_exception=False)
    if lineage_item is None or lineage_item.super_lineage_obj is None:
        logger.warning("Cannot update the lineage for tran job %s, because it does not exist.", train_id)
        raise ParamValueError("Cannot update the lineage for tran job %s, because it does not exist." % train_id)

    cached_added_info = lineage_item.super_lineage_obj.added_info
    new_added_info = dict(cached_added_info)

    for key, value in added_info.items():
        new_added_info.update({key: value})

    with cache_item.lock_key(LINEAGE):
        cache_item.get(key=LINEAGE).super_lineage_obj.added_info = new_added_info


class LineageCacheItemUpdater(BaseCacheItemUpdater):
    """Cache item updater for lineage info."""

    def update_item(self, cache_item: CachedTrainJob):
        """Update cache item in place."""
        summary_base_dir = cache_item.summary_base_dir
        summary_dir = cache_item.abs_summary_dir

        # The summary_base_dir and summary_dir have been normalized in data_manager.
        if summary_base_dir == summary_dir:
            relative_path = "./"
        else:
            relative_path = f'./{os.path.basename(summary_dir)}'

        try:
            lineage_parser = self._lineage_parsing(cache_item)
        except LineageFileNotFoundError:
            self._delete_lineage_in_cache(cache_item, LINEAGE, relative_path)
            return

        cache_item.set(key=LINEAGE, value=lineage_parser)

    def _lineage_parsing(self, cache_item):
        """Parse summaries and return lineage parser."""
        train_id = cache_item.train_id
        summary_dir = cache_item.abs_summary_dir
        update_time = cache_item.basic_info.update_time

        cached_lineage_item = cache_item.get(key=LINEAGE, raise_exception=False)
        if cached_lineage_item is None:
            lineage_parser = LineageParser(train_id, summary_dir, update_time)
        else:
            lineage_parser = cached_lineage_item
            with cache_item.lock_key(LINEAGE):
                lineage_parser.update_time = update_time
                lineage_parser.load()

        return lineage_parser

    def _delete_lineage_in_cache(self, cache_item, key, relative_path):
        with cache_item.lock_key(key):
            try:
                cache_item.delete(key=key)
                logger.info("Parse failed, delete the tran job %s.", relative_path)
            except ParamValueError:
                logger.debug("Parse failed, and it is not in cache, "
                             "no need to delete the train job %s.", relative_path)



# Copyright 2019 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""A reservoir sampling on the values."""

import random
import threading

from mindinsight.datavisual.common.log import logger
from mindinsight.datavisual.common.enums import PluginNameEnum
from mindinsight.utils.exceptions import ParamValueError
from mindinsight.datavisual.utils.utils import calc_histogram_bins


def binary_search(samples, target):
    """Binary search target in samples."""
    left = 0
    right = len(samples) - 1
    while left <= right:
        mid = (left + right) // 2
        if target < samples[mid].step:
            right = mid - 1
        elif target > samples[mid].step:
            left = mid + 1
        else:
            return mid

    # if right is -1, it is less than the first one.
    # if list is [1, 2, 4], target is 3, right will be 1, so wo will insert by 2.
    return right + 1


class Reservoir:
    """
    A container based on Reservoir Sampling algorithm.

    The newly added sample will be preserved. If the container is full, an old
    sample will be replaced randomly. The probability of each sample being
    replaced is the same.
    """

    def __init__(self, size):
        """
        A Container constructor which create a new Reservoir.

        Args:
            size (int): Container Size. If the size is 0, the container is not limited.

        Raises:
            ValueError: If size is negative integer.
        """
        if not isinstance(size, (int,)) or size < 0:
            raise ParamValueError('size must be nonnegative integer, was %s' % size)

        self._samples_max_size = size
        self._samples = []
        self._sample_counter = 0
        self._sample_selector = random.Random(0)
        self._mutex = threading.Lock()

    def samples(self):
        """Return all stored samples."""
        with self._mutex:
            return list(self._samples)

    def add_sample(self, sample):
        """
        Add a sample to Reservoir.

        Replace the old sample when the capacity is full.
        New added samples are guaranteed to be added to the reservoir.

        Args:
            sample (Any): The sample to add to the Reservoir.
        """
        with self._mutex:
            if len(self._samples) < self._samples_max_size or self._samples_max_size == 0:
                self._add_sample(sample)
            else:
                # Use the Reservoir Sampling algorithm to replace the old sample.
                rand_int = self._sample_selector.randint(0, self._sample_counter)
                if rand_int < self._samples_max_size:
                    self._samples.pop(rand_int)
                else:
                    self._samples = self._samples[:-1]
                self._add_sample(sample)
            self._sample_counter += 1

    def _add_sample(self, sample):
        """Search the index and add sample."""
        if not self._samples or sample.step > self._samples[-1].step:
            self._samples.append(sample)
            return
        index = binary_search(self._samples, sample.step)
        if index == len(self._samples):
            self._samples.append(sample)
        else:
            self._samples.insert(index, sample)

    def remove_sample(self, filter_fun):
        """
        Remove the samples from Reservoir that do not meet the filter criteria.

        Args:
            filter_fun (Callable[..., Any]): Determines whether a sample meets
                the deletion condition.

        Returns:
            int, the number of samples removed.
        """
        remove_size = 0

        with self._mutex:
            before_remove_size = len(self._samples)
            if before_remove_size > 0:
                # remove samples that meet the filter criteria.
                self._samples = list(filter(filter_fun, self._samples))
                after_remove_size = len(self._samples)
                remove_size = before_remove_size - after_remove_size

                if remove_size > 0:
                    # update _sample_counter when samples has been removed.
                    sample_remaining_rate = float(
                        after_remove_size) / before_remove_size
                    self._sample_counter = int(
                        round(self._sample_counter * sample_remaining_rate))

        return remove_size


class _VisualRange:
    """Simple helper class to merge visual ranges."""
    def __init__(self):
        self._max = 0.0
        self._min = 0.0
        self._updated = False

    def update(self, max_val: float, min_val: float) -> None:
        """
        Merge visual range with given range.

        Args:
            max_val (float): Max value of given range.
            min_val (float): Min value of given range.

        """
        if not self._updated:
            self._max = max_val
            self._min = min_val
            self._updated = True
            return

        if max_val > self._max:
            self._max = max_val

        if min_val < self._min:
            self._min = min_val

    @property
    def max(self):
        """Gets max value of current range."""
        return self._max

    @property
    def min(self):
        """Gets min value of current range."""
        return self._min


class HistogramReservoir(Reservoir):
    """
    Reservoir for histogram, which needs updating range over all steps.

    Args:
        size (int): Container Size. If the size is 0, the container is not limited.
    """
    def __init__(self, size):
        super().__init__(size)
        # Marker to avoid redundant calc for unchanged histograms.
        self._visual_range_up_to_date = False

    def add_sample(self, sample):
        """Adds sample, see parent class for details."""
        super().add_sample(sample)
        self._visual_range_up_to_date = False

    def samples(self):
        """Return all stored samples."""
        with self._mutex:
            if self._visual_range_up_to_date:
                return list(self._samples)

            # calc visual range
            visual_range = _VisualRange()
            max_count = 0
            for sample in self._samples:
                histogram_container = sample.value
                if histogram_container.count == 0:
                    # ignore empty tensor
                    continue
                max_count = max(histogram_container.count, max_count)
                visual_range.update(histogram_container.max, histogram_container.min)

            if visual_range.max == visual_range.min and not max_count:
                logger.debug("Max equals to min. Count is zero.")

            bins = calc_histogram_bins(max_count)

            # update visual range
            logger.debug(
                "Visual histogram: min %s, max %s, bins %s, max_count %s.",
                visual_range.min,
                visual_range.max,
                bins,
                max_count)
            for sample in self._samples:
                histogram = sample.value.histogram
                histogram.set_visual_range(visual_range.max, visual_range.min, bins)

            self._visual_range_up_to_date = True
            return list(self._samples)


class ReservoirFactory:
    """Factory class to get reservoir instances."""
    def create_reservoir(self, plugin_name: str, size: int) -> Reservoir:
        """
        Creates reservoir for given plugin name.

        Args:
            plugin_name (str): Plugin name
            size (int): Container Size. If the size is 0, the container is not limited.

        Returns:
            Reservoir, reservoir instance for given plugin name.
        """
        if plugin_name in (PluginNameEnum.HISTOGRAM.value, PluginNameEnum.TENSOR.value):
            return HistogramReservoir(size)
        return Reservoir(size)



# Copyright 2022-2022 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""The msadvisor proposer."""
import os
import json
import stat

from collections import OrderedDict

from mindinsight.profiler.proposer.allproposers.base_proposer import Proposer
from mindinsight.profiler.common.log import logger
from mindinsight.profiler.common.validator.validate_path import validate_and_normalize_path
from mindinsight.profiler.common.exceptions.exceptions import ProfilerDirNotFoundException


class MsadvisorProposer(Proposer):
    """The msadvisor proposer."""

    def __init__(self, profiling_dir, rank_id):
        super().__init__(profiling_dir, rank_id)
        self.__proposer_type = "msadvisor"
        self.__proposal_dict = OrderedDict()

    def analyze(self, options=None):
        """
        Get the proposal from msadvisor proposer.

        Args:
            options (dict): options for proposer analysis. Default: None.

        Returns:
            dict, the proposal from msadvisor proposer instance，the dictionary key is "msadvisor-proposer"
            including different models in msadvisor, such as AICPUModel, TransDataModel.

        Examples:
            >>> proposer_type = 'msadvisor'
            >>> proposer = MsadvisorProposer.instance().get_proposer(proposer_type, self.profiling_dir, self.device_id)
            >>> result = proposer.analyze(options)
        """
        try:
            logger.info("The MsadvisorProposer is running.")
            self.read_recommendation()
        except ProfilerDirNotFoundException as err:
            logger.warning(err)
        finally:
            pass

        return self.__proposal_dict

    def read_recommendation(self):
        """
        Get advice from msadvisor recommendation json file.

        Raises:
            ProfilerDirNotFoundException: If recommendation path does not exist.
        """
        recommendation_path = os.path.join(self.profiling_path, "msadvisor")
        recommendation_path = os.path.join(recommendation_path, "device_" + self.rank_id)
        recommendation_path = os.path.join(recommendation_path, "recommendation")
        if not os.path.exists(recommendation_path):
            raise ProfilerDirNotFoundException("Msadvisor recommendation dir does not exist!")
        file_list = os.listdir(recommendation_path)
        file_list.sort(key=lambda fn: os.path.getmtime(os.path.join(recommendation_path, fn)), reverse=True)
        for rec_file in file_list:
            if rec_file.endswith('.json'):
                recommendation_path = os.path.join(recommendation_path, rec_file)
                recommendation_path = validate_and_normalize_path(recommendation_path,
                                                                  'recommendation', allow_parent_dir=True)
                with os.fdopen(os.open(recommendation_path, os.O_RDONLY,
                                       stat.S_IRUSR | stat.S_IWUSR), "r") as recommendation_file:
                    recommendation = json.load(recommendation_file)
                    recommendation['msadvisor-proposer'] = recommendation.pop('model', None)
                    logger.info(recommendation)
                    self.__proposal_dict.update(recommendation)
                    logger.info("The MsadvisorProposer ran successfully.")
                break



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Data process analyser."""
import os

from mindinsight.profiler.analyser.base_analyser import BaseAnalyser
from mindinsight.profiler.common.validator.validate_path import validate_and_normalize_path

class MinddataAnalyser(BaseAnalyser):
    """The Minddata profiling analyser."""

    DEVICE_QUEUE_EMPTY_WARNING_THRESHOLD = 0.3

    def analyse_get_next_info(self, info_type="all"):
        """
        Analyse the get_next operation info.

        Args:
            info_type (str): The info type to return, default return both queue and time info,
            other options are ["queue", "time"].

        Returns:
            list[list], all get_next operation info, each info contains node_name, start, end, queue_size.
        """
        # init queue info result
        queue_info = dict()
        queue_size_list = []
        empty_step_count = 0

        # init time info result
        time_info = dict()
        time_list = []
        total_cost = 0

        file_name = "minddata_aicpu_" + self._device_id + ".txt"
        file_path = MinddataAnalyser.find_target_file(self._profiling_dir, file_name)

        # the GPU minddata profiler file
        if not file_path:
            file_name = "minddata_getnext_profiling_" + self._device_id + ".txt"
            file_path = MinddataAnalyser.find_target_file(self._profiling_dir, file_name)

        if file_path:
            file_path = validate_and_normalize_path(
                file_path, raise_key="Invalid minddata_getnext file path.")
            with open(file_path) as data_file:
                for line in data_file.readlines():
                    node_info = line.split()
                    # Ascend:GetNext_dequeue_wait GPU:GetNext
                    if node_info and node_info[0][0:7] == "GetNext":
                        # analyse target info type
                        if len(node_info) > 3 and info_type in ["all", "queue"]:
                            queue_size_list.append(int(node_info[3]))
                            if node_info[3] == '0':
                                empty_step_count += 1
                        if len(node_info) > 2 and info_type in ["all", "time"]:
                            one_step_cost_time = (float(node_info[2]) - float(node_info[1]))/1e3
                            # The time stamp in Ascend is μs but in GPU is ns.
                            if 'minddata_getnext_profiling' in file_name:
                                one_step_cost_time = one_step_cost_time/1e3
                            time_list.append(one_step_cost_time)
                            total_cost += one_step_cost_time
                if info_type in ["all", "queue"]:
                    queue_info["size"] = len(queue_size_list)
                    queue_info["info"] = {"queue": queue_size_list}
                    queue_info["summary"] = {
                        "queue_summary": {
                            "empty_queue": empty_step_count
                        }
                    }
                if len(node_info) > 2 and info_type in ["all", "time"]:
                    time_info["size"] = len(time_list)
                    time_info["info"] = {"get_next": time_list}
                    if time_info["size"]:
                        time_info["summary"] = {
                            "time_summary": {
                                "avg_cost": "0" if not time_list else str(total_cost / len(time_list))
                            }
                        }

        return queue_info, time_info

    def analyse_device_queue_info(self, info_type="all"):
        """
        Analyse the device_queue operation info.

        Args:
            info_type (str): The info type to return, default return both queue and time info,
            other options are ["queue", "time"].

        Returns:
            dict, queue size info.
            dict, time cost info.
        """
        # init queue info result
        queue_info = dict()
        get_time_list, push_time_list, total_time_list = [], [], []
        total_cost, total_push, total_get = 0, 0, 0

        # init time info result
        time_info = dict()
        queue_size_list = []
        empty_step, full_step = 0, 0

        file_path = self.get_device_queue_file_path()

        if file_path:
            file_path = validate_and_normalize_path(
                file_path, raise_key="Invalid device_queue file path.")
            with open(file_path) as data_file:
                for line in data_file.readlines():
                    op_info = line.split()
                    # time info
                    if op_info and op_info[0] == "0" and info_type in ["all", "time"]:
                        # sub_type: 0 get_time, 1 push time, 2 total time
                        # op_info: 2: step num 3: cost time
                        if op_info[1] == "0":
                            get_time_list.append([int(op_info[2]), float(op_info[3])])
                            total_get += float(op_info[3])
                        elif op_info[1] == "1":
                            push_time_list.append([int(op_info[2]), float(op_info[3])])
                            total_push += float(op_info[3])
                        elif op_info[1] == "2":
                            total_time_list.append([int(op_info[2]), float(op_info[3])])
                            total_cost += float(op_info[3])
                    elif op_info and op_info[0] == "1" and info_type in ["all", "queue"]:
                        queue_size_list.append([int(op_info[2]), int(op_info[3])])
                        if op_info[1] == op_info[3]:
                            full_step += 1
                        if op_info[3] == "0":
                            empty_step += 1
            if info_type in ["all", "time"]:
                total_time_list = MinddataAnalyser.sort_step(total_time_list)
                push_time_list = MinddataAnalyser.sort_step(push_time_list)
                get_time_list = MinddataAnalyser.sort_step(get_time_list)

                time_info["size"] = len(total_time_list)
                time_info["info"] = {"total_cost": total_time_list,
                                     "push_cost": push_time_list,
                                     "get_cost": get_time_list}
                if time_info["size"]:
                    time_info["summary"] = {"time_summary": {"avg_cost": total_cost/time_info["size"]}}
                    time_info["summary"]["time_summary"]["get_cost"] = total_get/time_info["size"]
                    time_info["summary"]["time_summary"]["push_cost"] = total_push/time_info["size"]

            if info_type in ["all", "queue"]:
                queue_size_list = MinddataAnalyser.sort_step(queue_size_list)

                queue_info["size"] = len(queue_size_list)
                queue_info["info"] = {"queue": queue_size_list}
                queue_info["summary"] = {"queue_summary": {"empty_queue": empty_step}}
                queue_info["summary"]["queue_summary"]["full_queue"] = full_step

        return queue_info, time_info

    def get_device_queue_file_path(self):
        """
        Get device queue file path.

        Returns:
            str, the file path.
        """
        device_queue_file_name = "device_queue_profiling_" + self._device_id + ".txt"
        device_queue_file_path = MinddataAnalyser.find_target_file(self._profiling_dir, device_queue_file_name)
        feed_file_name = "dataset_iterator_profiling_" + self._device_id + ".txt"
        feed_file_path = MinddataAnalyser.find_target_file(self._profiling_dir, feed_file_name)
        file_path = ""
        if device_queue_file_path:
            file_path = device_queue_file_path
        elif not device_queue_file_path and feed_file_path:
            file_path = feed_file_path

        return file_path

    @staticmethod
    def analyse_queue_summary(get_next_queue_info, device_queue_info):
        """
        Analyse the queue summary info.

        Args:
            get_next_queue_info (dict): the get_next queue info return by ananlyser.
            device_queue_info (dict): the device queue info return by ananlyser.

        Returns:
            dict, the summary of queue.
        """
        result = {}
        if get_next_queue_info and device_queue_info:
            result = {"data_process": {"status": "normal"},
                      "device_queue_warning": {"status": "normal"},
                      "data_transmission": {"status": "normal"},
                      "get_next": {"status": "normal"}}

            get_next_queue_empty_count = get_next_queue_info.get(
                "summary", {}).get("queue_summary", {}).get("empty_queue", 0)
            result["get_next_queue_info"] = {
                "summary": {
                    "empty_batch_count": get_next_queue_empty_count,
                    "total_batch": get_next_queue_info.get("size")
                }
            }

            device_queue_empty_count = device_queue_info.get(
                "summary", {}).get("queue_summary", {}).get("empty_queue", 0)
            device_queue_full_count = device_queue_info.get(
                "summary", {}).get("queue_summary", {}).get("full_queue", 0)

            result["device_queue_info"] = {"summary": {
                "empty_batch_count": device_queue_empty_count,
                "full_batch_count": device_queue_full_count,
                "total_batch": device_queue_info.get("size")}}

            # Adapt to the case that the first step data in the GPU is always empty
            if device_queue_empty_count > device_queue_info.get("size", 0)*\
                    MinddataAnalyser.DEVICE_QUEUE_EMPTY_WARNING_THRESHOLD:
                result["data_process"]["status"] = "warning"
            elif device_queue_empty_count <= device_queue_info.get("size", 0)*\
                    MinddataAnalyser.DEVICE_QUEUE_EMPTY_WARNING_THRESHOLD and get_next_queue_empty_count > 0:
                result["data_transmission"]["status"] = "warning"
            else:
                result["device_queue_warning"]["status"] = "warning"

        elif device_queue_info and not get_next_queue_info:
            result = {"data_process": {"status": "normal"},
                      "device_queue_warning": {"status": "normal"},
                      "fpbp": {"status": "normal"}}

            device_queue_empty_count = device_queue_info.get(
                "summary", {}).get("queue_summary", {}).get("empty_queue", 0)
            device_queue_full_count = device_queue_info.get(
                "summary", {}).get("queue_summary", {}).get("full_queue", 0)

            result["device_queue_info"] = {
                "summary": {
                    "empty_batch_count": device_queue_empty_count,
                    "full_batch_count": device_queue_full_count,
                    "total_batch": device_queue_info.get("size")
                }
            }

            if device_queue_empty_count > device_queue_info.get("size", 0)*\
                    MinddataAnalyser.DEVICE_QUEUE_EMPTY_WARNING_THRESHOLD:
                result["data_process"]["status"] = "warning"
            else:
                result["device_queue_warning"]["status"] = "warning"

        return result


    @staticmethod
    def sort_step(step_info_list):
        """
        Sorting the list by the first item and return the list of second item.

        Args:
            step_info_list (list): the step info, contains [step_num, info].

        Returns:
            list, the info list sorted by step.
        """
        step_info_list.sort(key=lambda x: x[0])
        result = []
        for item in step_info_list:
            result.append(item[1])
        return result

    @staticmethod
    def find_target_file(file_dir, file_name):
        """
        Find the target file in dir, and return the find file's abs path or "".

        Args:
            file_dir (str): The target file dir.
            file_name (str): The target file name.

        Returns:
            str, the abs file path.
        """
        target_file_path = os.path.join(file_dir, file_name)
        if not os.path.exists(target_file_path):
            return ""

        return target_file_path

    def _filter(self, filter_condition):
        """
        Filter the profiling data according to the filter condition.

        Args:
            filter_condition (dict): The filter condition.
        """

    def _load(self):
        """Load data according to the parsed profiling files."""



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""The converter between proto format event of lineage and dict."""


def organize_graph(graph_message):
    """
    Convert a dataset graph to its dict format.

    Args:
        graph_message (DatasetGraph): Graph event message.

    Returns:
        dict, dataset graph.
    """
    result = {}
    # update current dataset graph dict
    result.update(_organize_current_dataset(
        parameter=getattr(graph_message, 'parameter'),
        operations=getattr(graph_message, 'operations'),
        sampler=getattr(graph_message, 'sampler')
    ))
    # update children dataset graph dict
    result.update(
        _organize_children(getattr(graph_message, 'children'))
    )

    return result


def _organize_children(children_message):
    """
    Convert children message to its dict format.

    Args:
        children_message (list[DatasetGraph]): Children message.

    Returns:
        dict, children dict of dataset graph.
    """
    children_list = []
    children_dict = {'children': children_list}
    if children_message:
        for child_event in children_message:
            child_dict = {}
            # update current dataset to child
            child_dict.update(
                _organize_current_dataset(
                    parameter=getattr(child_event, 'parameter'),
                    operations=getattr(child_event, 'operations'),
                    sampler=getattr(child_event, 'sampler')
                )
            )
            # update child's children
            child_dict.update(
                _organize_children(getattr(child_event, 'children'))
            )
            children_list.append(child_dict)
        children_dict['children'] = children_list

    return children_dict


def _organize_current_dataset(parameter, operations, sampler):
    """
    Convert current dataset message to its dict format.

    Note:
        Current dataset message include parameter, operations,
        sampler message of dataset graph event.

    Args:
        parameter (OperationParameter): Parameter message.
        operations (Operation): Operations message.
        sampler (Operation): Sampler message.

    Returns:
        dict, current dataset.
    """
    current_dataset = {}
    if parameter:
        current_dataset.update(
            _organize_parameter(parameter)
        )
    if operations:
        operation_list = []
        for operation in operations:
            operation_list.append(
                _organize_operation(operation)
            )
        current_dataset.update(
            {'operations': operation_list}
        )
    if sampler:
        if _organize_operation(sampler):
            current_dataset.update({
                'sampler':
                    _organize_operation(sampler)
            })
    return current_dataset


def _organize_operation(operation):
    """
    Convert operation message to its dict format.

    Args:
        operation (Operation): Operation message.

    Returns:
        dict, operation.
    """
    operation_dict = {}
    operation_dict.update(_organize_parameter(getattr(operation, 'operationParam')))
    tmp_list = []
    repeated_keys = ['size', 'weights']
    for key in repeated_keys:
        for str_ele in getattr(operation, key):
            tmp_list.append(str_ele)
            dict()
        if tmp_list:
            operation_dict.update({key: tmp_list})
    return operation_dict


def _organize_parameter(parameter):
    """
    Convert operation parameter message to its dict format.

    Args:
        parameter (OperationParameter): Operation parameter message.

    Returns:
        dict, operation parameter.
    """
    parameter_result = dict()
    parameter_keys = [
        'mapStr',
        'mapBool',
        'mapInt',
        'mapDouble',
    ]
    for parameter_key in parameter_keys:
        base_attr = getattr(parameter, parameter_key)
        parameter_value = dict(base_attr)
        # convert str 'None' to None
        for key, value in parameter_value.items():
            if value == 'None':
                parameter_value[key] = None
        parameter_result.update(parameter_value)
    # drop `mapStrList` and `strValue` keys in result parameter
    str_list_para = dict(getattr(parameter, 'mapStrList'))
    result_str_list_para = dict()
    for key, value in str_list_para.items():
        str_list_para_list = list()
        for str_ele in getattr(value, 'strValue'):
            str_list_para_list.append(str_ele)
        str_list_para_list = list(map(lambda x: None if x == '' else x, str_list_para_list))
        result_str_list_para[key] = str_list_para_list
    parameter_result.update(result_str_list_para)

    return parameter_result



# Copyright 2020-2022 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Definition of error code and relative messages in debugger module."""
from mindinsight.utils.exceptions import MindInsightException
from mindinsight.debugger.common.exceptions.error_code import DebuggerErrors, DebuggerErrorMsg


class DebuggerParamTypeError(MindInsightException):
    """The parameter type error in debugger module."""

    def __init__(self, msg):
        super(DebuggerParamTypeError, self).__init__(
            error=DebuggerErrors.PARAM_TYPE_ERROR,
            message=DebuggerErrorMsg.PARAM_TYPE_ERROR.value.format(msg),
            http_code=400
        )


class DebuggerParamValueError(MindInsightException):
    """The parameter value error in debugger module."""

    def __init__(self, msg):
        super(DebuggerParamValueError, self).__init__(
            error=DebuggerErrors.PARAM_VALUE_ERROR,
            message=DebuggerErrorMsg.PARAM_VALUE_ERROR.value.format(msg),
            http_code=400
        )


class DebuggerCreateWatchPointError(MindInsightException):
    """The error about creating watch point."""

    def __init__(self, msg):
        super(DebuggerCreateWatchPointError, self).__init__(
            error=DebuggerErrors.CREATE_WATCHPOINT_ERROR,
            message=DebuggerErrorMsg.CREATE_WATCHPOINT_ERROR.value.format(msg),
            http_code=400
        )


class DebuggerUpdateWatchPointError(MindInsightException):
    """The error about updating watch point."""

    def __init__(self, msg):
        super(DebuggerUpdateWatchPointError, self).__init__(
            error=DebuggerErrors.UPDATE_WATCHPOINT_ERROR,
            message=DebuggerErrorMsg.UPDATE_WATCHPOINT_ERROR.value.format(msg),
            http_code=400
        )


class DebuggerDeleteWatchPointError(MindInsightException):
    """The error about deleting watch point."""

    def __init__(self, msg):
        super(DebuggerDeleteWatchPointError, self).__init__(
            error=DebuggerErrors.DELETE_WATCHPOINT_ERROR,
            message=DebuggerErrorMsg.DELETE_WATCHPOINT_ERROR.value.format(msg),
            http_code=400
        )


class DebuggerRecheckError(MindInsightException):
    """The error about deleting watch point."""

    def __init__(self, msg):
        super(DebuggerRecheckError, self).__init__(
            error=DebuggerErrors.RECHECK_ERROR,
            message=DebuggerErrorMsg.RECHECK_ERROR.value.format(msg),
            http_code=400
        )


class DebuggerCompareTensorError(MindInsightException):
    """The error about comparing tensors."""

    def __init__(self, msg):
        super(DebuggerCompareTensorError, self).__init__(
            error=DebuggerErrors.COMPARE_TENSOR_ERROR,
            message=msg,
            http_code=400
        )


class DebuggerContinueError(MindInsightException):
    """The error about continuing debugging."""
    def __init__(self, msg):
        super(DebuggerContinueError, self).__init__(
            error=DebuggerErrors.CONTINUE_ERROR,
            message=DebuggerErrorMsg.CONTINUE_ERROR.value.format(msg),
            http_code=400
        )


class DebuggerPauseError(MindInsightException):
    """The error about pausing debugging."""
    def __init__(self, msg):
        super(DebuggerPauseError, self).__init__(
            error=DebuggerErrors.PAUSE_ERROR,
            message=DebuggerErrorMsg.PAUSE_ERROR.value.format(msg),
            http_code=400
        )


class DebuggerNodeNotInGraphError(MindInsightException):
    """The node is not in the graph."""
    def __init__(self, node_name, node_type=None):
        if node_type is not None:
            err_msg = f"Cannot find the node in graph by the given name. node name: {node_name}, type: {node_type}."
        else:
            err_msg = f"Cannot find the node in graph by the given name. node name: {node_name}."
        super(DebuggerNodeNotInGraphError, self).__init__(
            error=DebuggerErrors.NODE_NOT_IN_GRAPH_ERROR,
            message=err_msg,
            http_code=400
        )


class DebuggerGraphNotExistError(MindInsightException):
    """The graph does not exist."""
    def __init__(self):
        super(DebuggerGraphNotExistError, self).__init__(
            error=DebuggerErrors.GRAPH_NOT_EXIST_ERROR,
            message=DebuggerErrorMsg.GRAPH_NOT_EXIST_ERROR.value,
            http_code=400
        )


class DebuggerStepNumError(MindInsightException):
    """The graph does not exist."""
    def __init__(self):
        super(DebuggerStepNumError, self).__init__(
            error=DebuggerErrors.STEP_NUM_ERROR,
            message="The type of step number should be int32.",
            http_code=400
        )


class DebuggerTensorGraphError(MindInsightException):
    """The error about comparing tensors."""

    def __init__(self):
        super(DebuggerTensorGraphError, self).__init__(
            error=DebuggerErrors.TENSOR_GRAPH_ERROR,
            message=DebuggerErrorMsg.TENSOR_GRAPH_ERROR.value,
            http_code=400
        )


class DebuggerTensorHitError(MindInsightException):
    """The error about comparing tensors."""

    def __init__(self):
        super(DebuggerTensorHitError, self).__init__(
            error=DebuggerErrors.TENSOR_HIT_ERROR,
            message=DebuggerErrorMsg.TENSOR_HIT_ERROR.value,
            http_code=400
        )


class DebuggerSetRecommendWatchpointsError(MindInsightException):
    """The set recommend watchpoints error in debugger module."""

    def __init__(self):
        super(DebuggerSetRecommendWatchpointsError, self).__init__(
            error=DebuggerErrors.SET_RECOMMEND_WATCHPOINT_ERROR,
            message=DebuggerErrorMsg.SET_RECOMMEND_WATCHPOINT_ERROR.value,
            http_code=400
        )


class DebuggerConditionUnavailableError(MindInsightException):
    """The condition unavailable error in debugger module."""

    def __init__(self, msg):
        super(DebuggerConditionUnavailableError, self).__init__(
            error=DebuggerErrors.DEBUGGER_CONDITION_UNAVAILABLE_ERROR,
            message=DebuggerErrorMsg.DEBUGGER_CONDITION_UNAVAILABLE_ERROR.value.format(msg),
            http_code=400
        )


class DebuggerServerRunningError(MindInsightException):
    """The server running error in debugger module."""

    def __init__(self, msg):
        super(DebuggerServerRunningError, self).__init__(
            error=DebuggerErrors.DEBUGGER_SERVER_RUNNING_ERROR,
            message=DebuggerErrorMsg.DEBUGGER_SERVER_RUNNING_ERROR.value.format(msg),
            http_code=500
        )


class DeviceIdUnregistered(MindInsightException):
    """The error of that the device id is unregister."""

    def __init__(self, msg):
        super(DeviceIdUnregistered, self).__init__(
            error=DebuggerErrors.DEVICE_ID_UNREGISTERED,
            message=DebuggerErrorMsg.DEVICE_ID_UNREGISTERED.value.format(msg),
            http_code=400
        )


class DebuggerModuleNotFoundError(MindInsightException):
    """The error of that the module is not found."""

    def __init__(self, msg):
        super(DebuggerModuleNotFoundError, self).__init__(
            error=DebuggerErrors.MODULE_NOT_FOUND_ERROR,
            message=DebuggerErrorMsg.MODULE_NOT_FOUND_ERROR.value.format(msg),
            http_code=500
        )


class DebuggerSessionNumOverBoundError(MindInsightException):
    """The error of that the session number is out of bound."""

    def __init__(self):
        super(DebuggerSessionNumOverBoundError, self).__init__(
            error=DebuggerErrors.DEBUGGER_SESSION_OVER_BOUND_ERROR,
            message=DebuggerErrorMsg.DEBUGGER_SESSION_OVER_BOUND_ERROR.value,
            http_code=400
        )


class DebuggerSessionNotFoundError(MindInsightException):
    """The error of that the session is not found."""

    def __init__(self, msg):
        super(DebuggerSessionNotFoundError, self).__init__(
            error=DebuggerErrors.DEBUGGER_SESSION_NOT_FOUND_ERROR,
            message=DebuggerErrorMsg.DEBUGGER_SESSION_NOT_FOUND_ERROR.value.format(msg),
            http_code=400
        )


class DebuggerOnlineSessionUnavailable(MindInsightException):
    """The error of that the online session is unavailable."""

    def __init__(self):
        super(DebuggerOnlineSessionUnavailable, self).__init__(
            error=DebuggerErrors.DEBUGGER_ONLINE_SESSION_UNAVAILABLE,
            message=DebuggerErrorMsg.DEBUGGER_ONLINE_SESSION_UNAVAILABLE.value,
            http_code=400
        )


class DebuggerDownloadOverQueue(MindInsightException):
    """The error of that the download queue is oversize."""

    def __init__(self):
        super(DebuggerDownloadOverQueue, self).__init__(
            error=DebuggerErrors.DEBUGGER_DOWNLOAD_OVER_QUEUE,
            message=DebuggerErrorMsg.DEBUGGER_DOWNLOAD_OVER_QUEUE.value,
            http_code=400
        )


class DebuggerDownloadTensorNotExist(MindInsightException):
    """The error of that the Tensor is not exist."""

    def __init__(self):
        super(DebuggerDownloadTensorNotExist, self).__init__(
            error=DebuggerErrors.DEBUGGER_DOWNLOAD_TENSOR_NOT_EXIST,
            message=DebuggerErrorMsg.DEBUGGER_DOWNLOAD_TENSOR_NOT_EXIST.value,
            http_code=400
        )


class RankDirNotFound(MindInsightException):
    """The error of that the dumped rank directory is not found."""

    def __init__(self, msg):
        super(RankDirNotFound, self).__init__(
            error=DebuggerErrors.DEBUGGER_RANK_DIR_NOT_FOUND,
            message=DebuggerErrorMsg.DEBUGGER_RANK_DIR_NOT_FOUND.value.format(msg),
            http_code=400
        )


class DebuggerJsonFileParseError(MindInsightException):
    """The error of that failed to parse the debugger json files."""

    def __init__(self, msg):
        super(DebuggerJsonFileParseError, self).__init__(
            error=DebuggerErrors.DEBUGGER_JSON_FILE_PARSE_ERROR,
            message=DebuggerErrorMsg.DEBUGGER_JSON_FILE_PARSE_ERROR.value.format(msg),
            http_code=400
        )


class DebuggerHistoryNotFoundError(MindInsightException):
    """The graph history and graph file doesn't match."""

    def __init__(self, msg):
        super(DebuggerHistoryNotFoundError, self).__init__(
            error=DebuggerErrors.DEBUGGER_HISTORY_NOT_FOUND_ERROR,
            message=DebuggerErrorMsg.DEBUGGER_HISTORY_NOT_FOUND_ERROR.value.format(msg),
            http_code=500
        )


class DebuggerHistoryValueError(MindInsightException):
    """The dumped step id is not belong to relative graph history."""

    def __init__(self, graph_id):
        super(DebuggerHistoryValueError, self).__init__(
            error=DebuggerErrors.DEBUGGER_STEP_VALUE_ERROR,
            message=DebuggerErrorMsg.DEBUGGER_STEP_VALUE_ERROR.value.format(graph_id),
            http_code=500
        )


class DebuggerToolkitNotFoundError(MindInsightException):
    """The error of that the module is not found."""

    def __init__(self, msg):
        super(DebuggerToolkitNotFoundError, self).__init__(
            error=DebuggerErrors.TOOLKIT_NOT_FOUND_ERROR,
            message=DebuggerErrorMsg.TOOLKIT_NOT_FOUND_ERROR.value.format(msg),
            http_code=500
        )


class DebuggerNodeTooLarge(MindInsightException):
    """The error of that the module is not found."""

    def __init__(self, limit, actual_val):
        super(DebuggerNodeTooLarge, self).__init__(
            error=DebuggerErrors.NODE_TOO_LARGE_ERROR,
            message=DebuggerErrorMsg.NODE_TOO_LARGE_ERROR.value.format(limit, actual_val),
            http_code=500
        )



# Copyright 2022 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""The Timeline Processor for Marey's Graph."""
import os
import re
import json
import collections
from mindinsight.profiler.common.util import analyse_device_list_from_profiler_dir
from mindinsight.profiler.common.validator.validate_path import validate_and_normalize_path


class TimelineService:
    """
    Analyse timeline data for marey's graph from file.
    """
    _ascend_display_filename = 'ascend_timeline_display_{}.json'
    _gpu_display_filename = 'gpu_timeline_display_{}.json'

    def __init__(self, path, device_list):
        self._display_timeline = True
        self.__read_data(path, device_list)
        self.__align_time()

    def get_ops_by_step(self, step):
        """
        Get the operators of the given step.

        Args:
            step (int): The current step.
        """
        min_time = float('inf')
        max_time = 0
        operator_time_maps = {}

        for device_name, cur_op_nodes in self.op_nodes.items():
            step_start = 0.0
            step_end = float('inf')
            for item in self.all_data.get(device_name):
                if item['name'] == step:
                    step_start = float(item['ts'])
                    step_end = step_start + float(item['dur'])
                    break

            operator_time_maps[device_name] = {}
            for item in cur_op_nodes:
                ts = float(item['ts'])
                dur = float(item['dur'])
                if step_start < ts < step_end or dur < step_start < ts + dur:
                    operator_time_maps.get(device_name)[item['name']] = {"st": ts,
                                                                         "ed": ts + dur,
                                                                         "dur": dur}
                    max_time = max(max_time, ts + dur)
                    min_time = min(min_time, ts)

        def cmp(a):
            return int(re.search(r'\d+', a[0]).group())

        a = 0
        stage_data = {}
        for device_names in sorted(self.stage_info, key=cmp):
            stage_name = 'stage' + str(a)
            a += 1
            data = {}
            for device_name in device_names:
                nodes = operator_time_maps.get(device_name)
                for node_name, node in nodes.items():
                    if node_name in data:
                        aggre_node = data.get(node_name)
                        aggre_node['st_max'] = max(aggre_node.get('st_max'), node.get('st'))
                        aggre_node['st_min'] = min(aggre_node.get('st_min'), node.get('st'))
                        aggre_node['ed_max'] = max(aggre_node.get('ed_max'), node.get('ed'))
                        aggre_node['ed_min'] = min(aggre_node.get('ed_min'), node.get('ed'))
                        aggre_node['st_avg'] = (node['st'] + aggre_node['st_avg'] *
                                                aggre_node['n']) / (aggre_node['n'] + 1)
                        aggre_node['ed_avg'] = (node['ed'] + aggre_node['ed_avg'] *
                                                aggre_node['n']) / (aggre_node['n'] + 1)
                        aggre_node['n'] += 1
                    else:
                        data[node_name] = {
                            'st_max': node['st'],
                            'st_min': node['st'],
                            'st_avg': node['st'],
                            'ed_max': node['ed'],
                            'ed_min': node['ed'],
                            'ed_avg': node['ed'],
                            'n': 1
                        }
            stage_data[stage_name] = {"data": data, "devices": device_names}
        TimelineData = collections.namedtuple('TimelineData',
                                              ['operator_time_maps', 'min_time', 'max_time', 'stage_data'])
        timeline_data = TimelineData(operator_time_maps, min_time, max_time, stage_data)
        return timeline_data

    def __read_data(self, path, device_list):
        """
        Get timeline data.

        Args:
            path (string): The current train log directory.
        """
        if device_list is None:
            device_list, _, _ = analyse_device_list_from_profiler_dir(path)
            sorted_device_list = sorted(device_list, key=int)
            device_list = sorted_device_list[:8]
        else:
            device_list = device_list.split(",")

        self.all_data = {}
        self.op_nodes = {}
        for device in device_list:
            display_filename = self._ascend_display_filename.format(device)
            device_entry = "device" + device
            file_path = os.path.join(path, display_filename)
            file_path = validate_and_normalize_path(
                file_path, raise_key='Invalid timeline display file path.'
            )
            if not os.path.exists(file_path):
                continue
            with open(file_path, "r") as fp:
                j_data = json.loads(fp.read())
                self.all_data[device_entry] = j_data
        for device_name, cur_data in self.all_data.items():
            self.op_nodes[device_name] = list(filter(_filt_op, cur_data))

    def __align_time(self):
        """Align timeline on each device."""
        self.stage_info = []
        if not self._display_timeline:
            return
        visited = set()
        self.align_info = {}
        one_step_op = self.__get_raw_ops_by_step("1")
        stages = []
        for device_name, cur_one_step_op in one_step_op.items():
            if device_name in visited:
                continue
            visited.add(device_name)
            stages.append(device_name)
            self.align_info[device_name] = 0
            min_all_reduce = ""
            minn = float('inf')
            for item in cur_one_step_op:
                if 'name' in item and 'AllReduce' in item['name']:
                    ts = float(item['ts'])
                    if ts < minn:
                        minn = ts
                        min_all_reduce = item
            if min_all_reduce == '':
                continue
            for device_name2 in self.all_data:
                if device_name2 in visited:
                    continue
                for item in one_step_op.get(device_name2):
                    if item['name'] == min_all_reduce['name']:
                        visited.add(device_name2)
                        min_all_reduce['ed'] = float(min_all_reduce['ts']) + float(min_all_reduce['dur'])
                        item['ed'] = float(item['ts']) + float(item['dur'])
                        self.align_info[device_name2] = min_all_reduce['ed'] - item['ed']
                        stages.append(device_name2)
                        break
            self.stage_info.append(stages)
            stages = []
        if stages:
            self.stage_info.append(stages)
        for device_name, cur_data in self.all_data.items():
            for item in cur_data:
                if 'ts' in item:
                    item['ts'] = float(item['ts']) + self.align_info.get(device_name)
        self.stage_device_map = {}

        def cmp(a):
            return int(re.search(r'\d+', a[0]).group())

        a = 0
        for device_names in sorted(self.stage_info, key=cmp):
            stage_name = 'stage' + str(a)
            a += 1
            self.stage_device_map[stage_name] = device_names

    def __get_raw_ops_by_step(self, step):
        """
        Get the raw operators of the given step.

        Args:
            step (int): The current step.
        """
        ret = {}
        for device_name, cur_op_nodes in self.op_nodes.items():
            ret[device_name] = []
            step_start = 0.0
            step_end = float('inf')
            for item in self.all_data.get(device_name):
                if item['name'] == step:
                    step_start = float(item['ts'])
                    step_end = step_start + float(item['dur'])
                    break
            for item in cur_op_nodes:
                if step_start < float(item['ts']) < step_end:
                    ret.get(device_name).append(item)
        return ret

    def __get_scopes_by_step(self, step):
        """
        Get the name scopes of the given step.

        Args:
            step (int): The current step.
        """
        ret = {}
        for device_name, cur_data in self.all_data.items():
            ret[device_name] = []
            step_start = 0.0
            step_end = float('inf')
            for item in cur_data:
                if item['name'] == step:
                    step_start = float(item['ts'])
                    step_end = step_start + float(item['dur'])
                    break
            for item in cur_data:
                if 'scope_level' in item:
                    if step_start < float(item['ts']) < step_end:
                        ret.get(device_name).append(item)
        return ret


def _filt_op(item):
    if 'AtomicAddrClean' in item['name'] or 'StreamSend' in item['name'] or 'StreamReceive' in item['name']:
        return False
    if 'tid' in item and '-op' in item['name'] and 'dur' in item and 'ts' in item:
        return True
    return False


def _find_scope(cur_scope_by_level, op):
    """Judge if op is in current scope."""
    for cur_scope in cur_scope_by_level:
        intervals = cur_scope_by_level[cur_scope]
        l = 0
        r = len(intervals) - 1
        ans = -1
        t = float(op['ts'])
        while l <= r:
            mid = (l + r) >> 1
            if intervals[mid][0] <= t:
                ans = max(ans, mid)
                l = mid + 1
            else:
                r = mid - 1
        if ans != -1 and intervals[ans][0] + intervals[ans][1] >= float(op['ts']) + float(op['dur']):
            return cur_scope, True
    return "", False



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""GenericNetwork module."""
import os
import textwrap

import click

from mindinsight.wizard.base.network import BaseNetwork
from mindinsight.wizard.base.templates import TemplateManager
from mindinsight.wizard.base.utility import process_prompt_choice, load_dataset_maker
from mindinsight.wizard.conf.constants import TEMPLATES_BASE_DIR
from mindinsight.wizard.conf.constants import QUESTION_START


class GenericNetwork(BaseNetwork):
    """BaseNetwork code generator."""
    name = 'GenericNetwork'
    supported_datasets = []
    supported_loss_functions = []
    supported_optimizers = []

    def __init__(self):
        self._dataset_maker = None
        template_dir = os.path.join(TEMPLATES_BASE_DIR, 'network', self.name.lower())
        self.network_template_manager = TemplateManager(os.path.join(template_dir, 'src'))
        self.common_template_manager = TemplateManager(template_dir, ['src', 'dataset'])

    def configure(self, settings=None):
        """
        Configure the network options.

        If settings is not None, then use the input settings to configure the network.

        Args:
            settings (dict): Settings to configure, format is {'options': value}.
                Example:
                    {
                        "loss": "SoftmaxCrossEntropyWithLogits",
                        "optimizer": "Momentum",
                        "dataset": "Cifar10"
                    }

        Returns:
            dict, configuration value to network.
        """
        if settings:
            config = dict(settings)
            dataset_name = settings['dataset']
            self._dataset_maker = load_dataset_maker(dataset_name)
        else:
            loss = self.ask_loss_function()
            optimizer = self.ask_optimizer()
            dataset_name = self.ask_dataset()
            self._dataset_maker = load_dataset_maker(dataset_name)
            dataset_config = self._dataset_maker.configure()

            config = {'loss': loss,
                      'optimizer': optimizer,
                      'dataset': dataset_name}
            config.update(dataset_config)
        self._dataset_maker.set_network(self)
        self.settings.update(config)
        return config

    @staticmethod
    def ask_choice(prompt_head, content_list, default_value=None):
        """Ask user to get selected result."""
        if default_value is None:
            default_choice = 1  # start from 1 in prompt message.
            default_value = content_list[default_choice - 1]

        choice_contents = content_list[:]
        choice_contents.sort(reverse=False)
        default_choice = choice_contents.index(default_value) + 1  # start from 1 in prompt message.

        prompt_msg = '{}:\n{}\n'.format(
            prompt_head,
            '\n'.join(f'{idx: >4}: {choice}' for idx, choice in enumerate(choice_contents, start=1))
        )
        prompt_type = click.IntRange(min=1, max=len(choice_contents))
        choice = click.prompt(prompt_msg, type=prompt_type, hide_input=False, show_choices=False,
                              confirmation_prompt=False, default=default_choice,
                              value_proc=lambda x: process_prompt_choice(x, prompt_type))
        click.secho(textwrap.dedent("Your choice is %s." % choice_contents[choice - 1]), fg='yellow')
        return choice_contents[choice - 1]

    def ask_loss_function(self):
        """Select loss function by user."""
        return self.ask_choice('%sPlease select a loss function' % QUESTION_START, self.supported_loss_functions)

    def ask_optimizer(self):
        """Select optimizer by user."""
        return self.ask_choice('%sPlease select an optimizer' % QUESTION_START, self.supported_optimizers)

    def ask_dataset(self):
        """Select dataset by user."""
        return self.ask_choice('%sPlease select a dataset' % QUESTION_START, self.supported_datasets)

    def generate(self, **options):
        """Generate network definition scripts."""
        context = self.get_generate_context(**options)
        network_source_files = self.network_template_manager.render(**context)
        for source_file in network_source_files:
            source_file.file_relative_path = os.path.join('src', source_file.file_relative_path)
        dataset_source_files = self._dataset_maker.generate(**options)
        for source_file in dataset_source_files:
            source_file.file_relative_path = os.path.join('src', source_file.file_relative_path)

        assemble_files = self._assemble(**options)
        source_files = network_source_files + dataset_source_files + assemble_files
        return source_files

    def get_generate_context(self, **options):
        """Get detailed info based on settings to network files."""
        context = dict(options)
        context.update(self.settings)
        return context

    def get_assemble_context(self, **options):
        """Get detailed info based on settings to assemble files."""
        context = dict(options)
        context.update(self.settings)
        return context

    def _assemble(self, **options):
        # generate train.py & eval.py & assemble scripts.
        assemble_files = []
        context = self.get_assemble_context(**options)
        common_source_files = self.common_template_manager.render(**context)
        assemble_files.extend(common_source_files)
        return assemble_files



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Hierarchical Occlusion encapsulator."""

from mindinsight.datavisual.common.exceptions import TrainJobNotExistError
from mindinsight.explainer.common.enums import ExplanationKeys, ImageQueryTypes
from mindinsight.explainer.encapsulator.explain_data_encap import ExplanationEncap


class HierarchicalOcclusionEncap(ExplanationEncap):
    """Hierarchical occlusion encapsulator."""

    def query_hierarchical_occlusion(self,
                                     train_id,
                                     labels,
                                     limit,
                                     offset,
                                     sorted_name,
                                     sorted_type,
                                     prediction_types=None,
                                     drop_empty=True,
                                     ):
        """
        Query hierarchical occlusion results.

        Args:
            train_id (str): Job ID.
            labels (list[str]): Label filter.
            limit (int): Maximum number of items to be returned.
            offset (int): Page offset.
            sorted_name (str): Field to be sorted.
            sorted_type (str): Sorting order, 'ascending' or 'descending'.
            prediction_types (list[str]): Prediction types filter.
            drop_empty (bool): Whether to drop out the data without hoc data. Default: True.

        Returns:
            tuple[int, list[dict]], total number of samples after filtering and list of sample results.
        """
        job = self.job_manager.get_job(train_id)
        if job is None:
            raise TrainJobNotExistError(train_id)

        if drop_empty:
            samples = self._query_samples(job, labels, sorted_name, sorted_type, prediction_types,
                                          drop_type=ExplanationKeys.HOC.value)
        else:
            samples = self._query_samples(job, labels, sorted_name, sorted_type, prediction_types)

        sample_infos = []
        obj_offset = offset * limit
        count = len(samples)
        end = count
        if obj_offset + limit < end:
            end = obj_offset + limit
        for i in range(obj_offset, end):
            sample = samples[i]
            sample_infos.append(self._touch_sample(sample, job, drop_empty))

        return count, sample_infos

    def _touch_sample(self, sample, job, drop_empty):
        """
        Final edit on single sample info.

        Args:
             sample (dict): Sample info.
             job (ExplainManager): Explain job.
             drop_empty (bool): Whether to drop out inferences without HOC explanations.

        Returns:
            dict, the edited sample info.
        """
        original = ImageQueryTypes.ORIGINAL.value
        outcome = ImageQueryTypes.OUTCOME.value

        sample["image"] = self._get_image_url(job.train_id, sample["image"], original)
        inferences = sample["inferences"]
        i = 0  # init index for while loop
        while i < len(inferences):
            inference_item = inferences[i]
            if drop_empty and not inference_item[ExplanationKeys.HOC.value]:
                inferences.pop(i)
                continue
            new_list = []
            for idx, hoc_layer in enumerate(inference_item[ExplanationKeys.HOC.value]):
                hoc_layer[outcome] = self._get_image_url(job.train_id,
                                                         f"{sample['id']}_{inference_item['label']}_{idx}.jpg",
                                                         outcome)
                new_list.append(hoc_layer)
            inference_item[ExplanationKeys.HOC.value] = new_list
            i += 1
        return sample



# Copyright 2019 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Scalar Processor APIs."""
from mindinsight.utils.exceptions import ParamValueError
from mindinsight.datavisual.common.log import logger
from mindinsight.datavisual.utils.tools import if_nan_inf_to_none
from mindinsight.datavisual.common.exceptions import ScalarNotExistError
from mindinsight.datavisual.common.exceptions import TrainJobNotExistError
from mindinsight.datavisual.common.validation import Validation
from mindinsight.datavisual.processors.base_processor import BaseProcessor


class ScalarsProcessor(BaseProcessor):
    """Scalar Processor."""

    def get_metadata_list(self, train_id, tag):
        """
        Builds a JSON-serializable object with information about scalars.

        Args:
            train_id (str): The ID of the events data.
            tag (str): The name of the tag the scalars all belonging to.

        Returns:
            list[dict], a list of dictionaries containing the `wall_time`, `step`, `value` for each scalar.
        """
        Validation.check_param_empty(train_id=train_id, tag=tag)
        job_response = []
        try:
            tensors = self._data_manager.list_tensors(train_id, tag)
        except ParamValueError as ex:
            raise ScalarNotExistError(ex.message)

        for tensor in tensors:
            job_response.append({
                'wall_time': tensor.wall_time,
                'step': tensor.step,
                'value': tensor.value})
        return dict(metadatas=job_response)

    def get_scalars(self, train_ids, tags):
        """
        Get scalar data for given train_ids and tags.

        Args:
            train_ids (list): Specify list of train job ID.
            tags (list): Specify list of tags.

        Returns:
            list[dict], a list of dictionaries containing the `wall_time`, `step`, `value` for each scalar.
        """
        scalars = []
        for train_id in train_ids:
            scalars += self._get_train_scalars(train_id, tags)

        return scalars

    def _get_train_scalars(self, train_id, tags):
        """
        Get scalar data for given train_id and tags.

        Args:
            train_id (str): Specify train job ID.
            tags (list): Specify list of tags.

        Returns:
            list[dict], a list of dictionaries containing the `wall_time`, `step`, `value` for each scalar.
        """
        scalars = []
        for tag in tags:
            try:
                tensors = self._data_manager.list_tensors(train_id, tag)
            except ParamValueError:
                continue
            except TrainJobNotExistError:
                logger.warning('Can not find the given train job in cache.')
                return []

            scalar = {
                'train_id': train_id,
                'tag': tag,
                'values': [],
            }

            for tensor in tensors:
                scalar['values'].append({
                    'wall_time': tensor.wall_time,
                    'step': tensor.step,
                    'value': if_nan_inf_to_none('scalar_value', tensor.value),
                })

            scalars.append(scalar)

        return scalars



# Copyright 2020-2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Implement the debugger grpc server."""
import copy
import threading
from functools import wraps
from queue import Queue, Full, Empty

import mindinsight
from mindinsight.common.util import version_match
from mindinsight.debugger.common.log import LOGGER as log
from mindinsight.debugger.common.utils import MAX_SINGLE_TENSOR_CACHE_BYTES, get_ack_reply, ServerStatus, \
    Streams, RunLevel, load_tensor, convert_tensor_stats, \
    put_tensor_stats_in_cache, put_tensor_base_in_cache, get_tensor_value
from mindinsight.debugger.conditionmgr.condition import TargetTypeEnum, ParamNameEnum
from mindinsight.debugger.debugger_services.debugger_server_base import debugger_server_wrap
from mindinsight.debugger.proto import debug_grpc_pb2_grpc as grpc_server_base
from mindinsight.debugger.proto.debug_grpc_pb2 import ViewCMD
from mindinsight.debugger.stream_operator.watchpoint_operator import WatchpointOperator
from mindinsight.domain.graph.proto.ms_graph_pb2 import GraphProto


def debugger_wrap(func):
    """Wrapper for catch exception."""

    @wraps(func)
    def record_log(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as err:
            log.exception(err)
            raise err

    return record_log


class DebuggerGrpcServer(grpc_server_base.EventListenerServicer):
    """The grpc server used to interactive with grpc client."""

    def __init__(self, cache_store):
        """
        Initialize.

        Args:
            cache_store (DebuggerCache): Debugger cache store.
        """
        cache_store.initialize()
        self._cache_store = cache_store
        # the next position of command queue to be queried
        self._pos = None
        # the status of grpc server, the value is in ServerStatus
        self._status = None
        # the run command cache, used to deal with left continue steps or nodes
        self._old_run_cmd = None
        # the view command cache, used to update tensor history through data queue
        self._received_view_cmd = None
        # the flag of receiving watch point hit
        self._received_hit = None
        self._heartbeat = None
        self.init()

    def init(self):
        """Init debugger grpc server."""
        self._pos = '0'
        self._status = ServerStatus.PENDING
        self._old_run_cmd = {}
        self._received_view_cmd = {}
        self._received_hit = []
        if self._heartbeat is not None:
            self._heartbeat.stop()
        self._heartbeat = None
        self._cache_store.clean()

    @debugger_wrap
    def WaitCMD(self, request, context):
        """Wait for a command in DebuggerCache."""
        # check if graph have already received.
        log.info("Received WaitCMD at %s-th step.", request.cur_step)
        if self._status == ServerStatus.PENDING:
            log.warning("No graph received before WaitCMD.")
            reply = get_ack_reply(1)
            return reply

        # send graph if it has not been sent before
        self._pre_process(request)
        # deal with old command
        reply = self._deal_with_old_command()
        # wait for next command
        if reply is None:
            reply = self._wait_for_next_command()
        # check the reply
        if reply is None:
            reply = get_ack_reply(1)
            log.warning("Failed to get command event.")
        else:
            log.debug("Reply to WaitCMD: %s", reply)
        return reply

    def _pre_process(self, request):
        """Pre-process before dealing with command."""

        # check if version is mismatch, if mismatch, send mismatch info to UI
        if self._status == ServerStatus.MISMATCH:
            log.warning("Version of MindSpore and MindInsight are unmatched,"
                        "waiting for user to terminate the script.")
            metadata_stream = self._cache_store.get_stream_handler(Streams.METADATA)
            # put metadata into data queue
            metadata = metadata_stream.get(['state', 'debugger_version'])
            self._cache_store.put_data(metadata)
            return

        metadata_stream = self._cache_store.get_stream_handler(Streams.METADATA)
        is_new_step = metadata_stream.step < request.cur_step
        is_new_node = metadata_stream.full_name != request.cur_node
        # clean cache data at the beginning of new step or node has been changed.
        if is_new_step or is_new_node:
            self._cache_store.clean_data()
            self._cache_store.get_stream_handler(Streams.TENSOR).get_tensor_handler_by_rank_id(0).clean_tensors(
                request.cur_step)
        if is_new_step:
            self._cache_store.get_stream_handler(Streams.WATCHPOINT_HIT).get_hit_handler_by_rank_id(0).clean()
        # receive graph at the beginning of the training
        if self._status == ServerStatus.RECEIVE_GRAPH:
            self._send_graph_flag(metadata_stream)
        # receive new metadata
        if is_new_step or is_new_node:
            self._update_metadata(metadata_stream, request)
        self._send_received_tensor_tag()
        self._send_watchpoint_hit_flag()

    def _send_graph_flag(self, metadata_stream):
        """
        Send graph and metadata to UI.

        Args:
            metadata_stream (MetadataHandler): Metadata handler stream.
        """
        self._cache_store.clean_command()
        # receive graph in the beginning of the training
        self._status = ServerStatus.WAITING
        metadata_stream.state = ServerStatus.WAITING.value
        metadata = metadata_stream.get()
        res = self._cache_store.get_stream_handler(Streams.GRAPH).get_graph_handler_by_rank_id(0).get()
        res.update(metadata)
        self._cache_store.put_data(res)
        log.debug("Put graph into data queue.")

    def _update_metadata(self, metadata_stream, metadata_proto):
        """
        Update metadata.

        Args:
            metadata_stream (MetadataHandler): Metadata handler stream.
            metadata_proto (MetadataProto): Metadata proto send by client.
        """
        # put new metadata into cache
        metadata_stream.put(metadata_proto)
        # update current node name and graph name
        graph_stream = self._cache_store.get_stream_handler(Streams.GRAPH).get_graph_handler_by_rank_id(0)
        full_name = metadata_proto.cur_node
        graph_name = graph_stream.get_graph_id_by_full_name(
            full_name) if full_name else metadata_stream.graph_name
        cur_node = graph_stream.get_node_name_by_full_name(full_name, graph_name)
        metadata_stream.node_name = cur_node
        metadata_stream.graph_name = graph_name
        metadata = metadata_stream.get()
        self._cache_store.put_data(metadata)
        log.debug("Put new metadata into data queue.")

    def _send_received_tensor_tag(self):
        """Send received_finish_tag."""
        node_info = self._received_view_cmd.get('node_info')
        if not node_info or self._received_view_cmd.get('wait_for_tensor'):
            return
        metadata = self._cache_store.get_stream_handler(Streams.METADATA).get(['step', 'state'])
        ret = {'receive_tensor': node_info.copy()}
        ret.update(metadata)
        self._cache_store.put_data(ret)
        self._received_view_cmd.clear()
        log.debug("Send receive tensor flag for %s", node_info)

    def _send_watchpoint_hit_flag(self):
        """Send Watchpoint hit flag."""
        watchpoint_hit_stream = self._cache_store.get_stream_handler(Streams.WATCHPOINT_HIT).get_hit_handler_by_rank_id(
            0)
        if not self._received_hit:
            return
        watchpoint_hits = self._received_hit
        self._received_hit = []
        for watchpoint_hit in watchpoint_hits:
            watchpoint_hit_stream.put(watchpoint_hit)
        watchpoint_hits_info = {'receive_watchpoint_hits': True}
        self._cache_store.put_data(watchpoint_hits_info)
        log.debug("Send the watchpoint hits to DataQueue.\nSend the reply.")

    def _deal_with_old_command(self):
        """Deal with old command."""
        event = None
        while self._cache_store.has_command(self._pos) and event is None:
            event = self._get_next_command()
            log.debug("Deal with old %s-th command:\n%s.", self._pos, event)
        # deal with continue run command
        if event is None and self._old_run_cmd:
            left_step_count = self._old_run_cmd.get('left_step_count')
            node_name = self._old_run_cmd.get('node_name')
            # node_name and left_step_count should not set at the same time
            if not (left_step_count or node_name) or (left_step_count and node_name):
                log.warning("Invalid old run command. %s", self._old_run_cmd)
                self._old_run_cmd.clear()
                return None
            if left_step_count:
                event = self._deal_with_left_continue_step(left_step_count)
            else:
                event = self._deal_with_left_continue_node(node_name)
            log.debug("Send old RunCMD.")
        return event

    def _deal_with_left_continue_step(self, left_step_count):
        """
        Construct run command with left continue step count.

        Args:
            left_step_count (int): The count of left steps to be executed.

        Returns:
            Event, the run command event.
        """
        event = get_ack_reply()
        event.run_cmd.run_steps = 1
        event.run_cmd.run_level = 'step'
        left_step_count = left_step_count - 1 if left_step_count > 0 else -1
        if not left_step_count:
            self._old_run_cmd.clear()
        else:
            self._old_run_cmd['left_step_count'] = left_step_count
        log.debug("Send old step RunCMD. Left step count: %s", left_step_count)
        return event

    def _deal_with_left_continue_node(self, node_name):
        """
        Construct run command with left continue nodes.

        Args:
            node_name (str): The target node name.

        Returns:
            Union[None, Event], the run command event.
        """
        cur_full_name = self._cache_store.get_stream_handler(Streams.METADATA).full_name
        if cur_full_name == node_name:
            log.info("Execute to target node: %s", node_name)
            self._old_run_cmd.clear()
            return None
        event = get_ack_reply()
        event.run_cmd.run_level = 'node'
        event.run_cmd.node_name = ''
        log.debug("Send old node RunCMD, cur node: %s, target node: %s", cur_full_name, node_name)
        return event

    def _wait_for_next_command(self):
        """
        Wait for next command.

        Returns:
            EventReply, the command event.
        """
        log.info("Start to wait for command.")
        if self._status != ServerStatus.MISMATCH:
            metadata_stream = self._cache_store.get_stream_handler(Streams.METADATA)
            metadata_stream.state = ServerStatus.WAITING.value
            self._cache_store.put_data(metadata_stream.get())
        event = None
        while event is None and self._status not in [ServerStatus.RUNNING, ServerStatus.PENDING]:
            log.debug("Wait for %s-th command", self._pos)
            event = self._get_next_command()
        return event

    def _get_next_command(self):
        """Get next command."""
        self._pos, event = self._cache_store.get_command(self._pos)
        if event is None:
            return event
        # deal with command
        metadata_stream = self._cache_store.get_stream_handler(Streams.METADATA)
        if isinstance(event, dict):
            event = self._deal_with_view_cmd(event)
        elif event.HasField('run_cmd'):
            event = self._deal_with_run_cmd(event)
            self._cache_store.put_data(metadata_stream.get())
        elif event.HasField('exit'):
            self._cache_store.clean()
            self._cache_store.put_data(metadata_stream.get())
            log.debug("Clean cache for exit cmd.")
        else:
            event = self._deal_with_set_cmd(event)
            log.debug("get set cmd.")

        return event

    def _deal_with_view_cmd(self, event):
        """
        Deal with view cmd.

        Args:
            event (dict): View command params.

                - view_cmd (EventReply): EventReply with view command.
                - node_name (str): The center node name for view command.
                - tensor_name (str): The center tensor name for view command.
                - graph_name (str): The graph name of center node.

        Returns:
            EventReply, view command to be sent to client.
        """
        view_cmd = event.pop('view_cmd', None)
        log.debug("Receive view cmd for node: %s.", event)
        if not (view_cmd and event):
            log.debug("Invalid view command. Ignore it.")
            return None
        self._received_view_cmd['node_info'] = event
        self._received_view_cmd['wait_for_tensor'] = True
        self._received_view_cmd['view_cmd'] = view_cmd
        level_dict = {
            'stats': ViewCMD.Level.statistics,
            'base': ViewCMD.Level.base
        }
        if event.get('level'):
            view_cmd.view_cmd.level = level_dict.get(event.get('level'))
        return view_cmd

    def _deal_with_run_cmd(self, event):
        """Deal with run cmd."""
        metadata_stream = self._cache_store.get_stream_handler(Streams.METADATA)
        run_cmd = event.run_cmd
        # receive step command
        if run_cmd.run_level == RunLevel.STEP.value:
            # receive pause cmd
            if not run_cmd.run_steps:
                log.debug("Pause training and wait for next command.")
                self._old_run_cmd.clear()
                # update metadata state from sending to waiting
                metadata_stream.state = ServerStatus.WAITING.value
                return None
            # receive step cmd
            left_steps = run_cmd.run_steps - 1
            event.run_cmd.run_steps = 1
            if left_steps:
                self._old_run_cmd['left_step_count'] = left_steps if left_steps > 0 else -1
        elif run_cmd.node_name:
            self._old_run_cmd['node_name'] = run_cmd.node_name
            run_cmd.node_name = ''
        # clean watchpoint hit cache
        if run_cmd.run_level == RunLevel.RECHECK.value:
            self._cache_store.get_stream_handler(Streams.WATCHPOINT_HIT).get_hit_handler_by_rank_id(0).clean()
            log.debug("Receive RunCMD. Clean watchpoint hit cache.")
        # update metadata state from sending to running
        metadata_stream.state = ServerStatus.RUNNING.value
        return event

    def _deal_with_set_cmd(self, event):
        """
        Deal with set cmd.

        Args:
            event (EventReply): EventReply with set command.

        Returns:
            EventReply, set command to be sent to client.
        """
        set_cmd = event.set_cmd
        watchpoint_stream = self._cache_store.get_stream_handler(Streams.WATCHPOINT)
        watchpoint_stream.clean_cache_set_cmd(set_cmd)
        if set_cmd.delete:
            return event
        # add watch node and condition info into set_cmd
        multi_card_graph_stream = self._cache_store.get_stream_handler(Streams.GRAPH)
        WatchpointOperator.add_set_cmd_detail(set_cmd, watchpoint_stream, multi_card_graph_stream)
        return event

    @debugger_wrap
    def SendMetadata(self, request, context):
        """Send metadata into DebuggerCache."""
        log.info("Received Metadata.")
        if self._status != ServerStatus.PENDING:
            log.info("Re-initialize cache store when new session comes.")
            self.init()

        client_ip = context.peer().split(':', 1)[-1]
        metadata_stream = self._cache_store.get_stream_handler(Streams.METADATA)
        reply = get_ack_reply()
        if request.training_done:
            log.info("The training from %s has finished.", client_ip)
        else:
            ms_version = request.ms_version
            version_match_stat = version_match(ms_version, mindinsight.__version__)
            if version_match_stat > 1:
                log.warning("Version is mismatched, MindSpore is: %s, MindInsight is: %s",
                            ms_version, mindinsight.__version__)
                self._status = ServerStatus.MISMATCH
                reply.version_matched = False
                metadata_stream.state = ServerStatus.MISMATCH.value
            elif version_match_stat > 0:
                log.warning("Version is not completely matched, MindSpore is: %s, MindInsight is: %s",
                            ms_version, mindinsight.__version__)
                reply.version_matched = True
            else:
                log.info("version is matched.")
                reply.version_matched = True

            metadata_stream.debugger_version = {'ms': ms_version, 'mi': mindinsight.__version__}
            log.debug("Put ms_version from %s into cache.", client_ip)

            metadata_stream.put(request)
            metadata_stream.client_ip = client_ip
            log.debug("Put new metadata from %s into cache.", client_ip)
        # put metadata into data queue
        metadata = metadata_stream.get()
        self._cache_store.put_data(metadata)
        log.debug("Send the reply to %s.", client_ip)
        return reply

    @debugger_wrap
    def SendGraph(self, request_iterator, context):
        """Send graph into DebuggerCache."""
        log.info("Received graph.")
        if self._old_run_cmd:
            log.info("Clear old run command. %s", self._old_run_cmd)
            self._old_run_cmd.clear()
        reply = get_ack_reply()
        if self._status == ServerStatus.MISMATCH:
            log.info("MindSpore and MindInsight is unmatched, waiting for user to terminate the service.")
            return reply
        serial_graph = b""
        for chunk in request_iterator:
            serial_graph += chunk.buffer
        graph = GraphProto.FromString(serial_graph)
        log.debug("Deserialize the graph %s. Receive %s nodes", graph.name, len(graph.node))
        graph_dict = {graph.name: graph}
        self._cache_store.get_stream_handler(Streams.GRAPH).put({0: graph_dict})
        self._cache_store.get_stream_handler(Streams.GRAPH).parse_stack_infos()
        self._cache_store.get_stream_handler(Streams.TENSOR).get_tensor_handler_by_rank_id(0).put_const_vals(
            graph.const_vals)
        self._cache_store.get_stream_handler(Streams.METADATA).graph_name = graph.name
        self._record_parameter_names()
        self._status = ServerStatus.RECEIVE_GRAPH
        log.debug("Send the reply for graph.")
        return reply

    @debugger_wrap
    def SendMultiGraphs(self, request_iterator, context):
        """Send graph into DebuggerCache."""
        log.info("Received multi_graphs.")
        if self._old_run_cmd:
            log.info("Clear old run command. %s", self._old_run_cmd)
            self._old_run_cmd.clear()
        reply = get_ack_reply()
        if self._status == ServerStatus.MISMATCH:
            log.info("MindSpore and MindInsight is unmatched, waiting for user to terminate the service.")
            return reply
        serial_graph = b""
        graph_dict = {}
        for chunk in request_iterator:
            serial_graph += chunk.buffer
            if chunk.finished:
                sub_graph = GraphProto.FromString(serial_graph)
                graph_dict[sub_graph.name] = sub_graph
                log.debug("Deserialize the graph %s. Receive %s nodes", sub_graph.name,
                          len(sub_graph.node))
                serial_graph = b""
                self._cache_store.get_stream_handler(Streams.TENSOR).get_tensor_handler_by_rank_id(0).put_const_vals(
                    sub_graph.const_vals)

        self._cache_store.get_stream_handler(Streams.GRAPH).put({0: graph_dict})
        self._cache_store.get_stream_handler(Streams.GRAPH).parse_stack_infos()
        self._record_parameter_names()
        self._status = ServerStatus.RECEIVE_GRAPH
        log.debug("Send the reply for graph.")
        return reply

    def _record_parameter_names(self):
        """Record parameter full names in tensor handler."""
        parameter_nodes = self._cache_store.get_stream_handler(Streams.GRAPH).get_graph_handler_by_rank_id(0) \
            .search_in_graph(pattern={'node_category': TargetTypeEnum.PARAMETER.value})
        tensor_stream = self._cache_store.get_stream_handler(Streams.TENSOR).get_tensor_handler_by_rank_id(0)
        for node in parameter_nodes:
            tensor_name = [node.full_name + ':0']
            tensor_stream.record_parameter_names(tensor_name)

    @debugger_wrap
    def SendTensors(self, request_iterator, context):
        """Send tensors into DebuggerCache."""
        log.info("Received tensor.")
        tensor_contents = []
        data_size = 0
        metadata_stream = self._cache_store.get_stream_handler(Streams.METADATA)
        tensor_stream = self._cache_store.get_stream_handler(Streams.TENSOR).get_tensor_handler_by_rank_id(0)
        step = metadata_stream.step
        oversize = False
        node_info = self._received_view_cmd.get('node_info')
        if node_info and node_info.get('load'):
            load_info = node_info.get('load')
            load_info['graph_name'] = node_info.get('graph_name')
            load_info['node_name'] = node_info.get('node_name')
            # The rank_id of online debugger is 0
            load_tensor(load_info=load_info, step=step, request_iterator=request_iterator,
                        cache_store=self._cache_store, rank_id=0)
        else:
            for tensor_proto in request_iterator:
                tensor_contents.append(tensor_proto.tensor_content)
                data_size += len(tensor_proto.tensor_content)
                if data_size >= MAX_SINGLE_TENSOR_CACHE_BYTES or oversize:
                    oversize = True
                    tensor_contents = []
                if tensor_proto.finished:
                    value = get_tensor_value(tensor_proto, tensor_contents, node_info, step, oversize, data_size)
                    update_flag = tensor_stream.put(value)
                    self._update_state(update_flag)
                    tensor_contents = []
                    data_size = 0
                    oversize = False
        reply = get_ack_reply()
        return reply

    @debugger_wrap
    def SendTensorBase(self, request_iterator, context):
        """Send tensor base into DebuggerCache."""
        log.info("Received tensor base.")
        metadata_stream = self._cache_store.get_stream_handler(Streams.METADATA)
        cur_step = metadata_stream.step
        tensor_stream = self._cache_store.get_stream_handler(Streams.TENSOR).get_tensor_handler_by_rank_id(0)
        tensor_protos = self._received_view_cmd.get('view_cmd').view_cmd.tensors
        tensor_stats_list = []
        for tensor_base in request_iterator:
            tensor_stats_list.append(_convert_tensor_base(tensor_base))
        update_data_flag = put_tensor_base_in_cache(tensor_stats_list, tensor_protos, cur_step, tensor_stream)
        self._update_state(update_data_flag)
        reply = get_ack_reply()
        return reply

    @debugger_wrap
    def SendTensorStats(self, request_iterator, context):
        """Send tensor stats into DebuggerCache."""
        log.info("Received tensor statistic.")
        metadata_stream = self._cache_store.get_stream_handler(Streams.METADATA)
        cur_step = metadata_stream.step
        tensor_stream = self._cache_store.get_stream_handler(Streams.TENSOR).get_tensor_handler_by_rank_id(0)
        tensor_protos = self._received_view_cmd.get('view_cmd').view_cmd.tensors
        tensor_stats_list = []
        for tensor_summary in request_iterator:
            tensor_stats_list.append({
                'tensor_base': _convert_tensor_base(tensor_summary.tensor_base),
                'tensor_stats': convert_tensor_stats(tensor_summary.statistics)
            })
        update_data_flag = put_tensor_stats_in_cache(tensor_stats_list, tensor_protos, cur_step, tensor_stream)
        self._update_state(update_data_flag)
        reply = get_ack_reply()
        return reply

    def _update_state(self, update_flag):
        """Update wait_for_tensor state."""
        if self._received_view_cmd.get('wait_for_tensor') and update_flag:
            # update_flag is used to avoid querying empty tensors again
            self._received_view_cmd['wait_for_tensor'] = False
            log.debug("Set wait for tensor flag to False.")

    @debugger_wrap
    def SendWatchpointHits(self, request_iterator, context):
        """Send watchpoint hits info DebuggerCache."""
        log.info("Received WatchpointHits. Left run cmd %s change to empty.", self._old_run_cmd)
        self._old_run_cmd.clear()
        if self._cache_store.get_stream_handler(Streams.METADATA).state == ServerStatus.RUNNING.value:
            # if the client session is running a script, all the cached command should be cleared
            # when received watchpoint_hits.
            self._cache_store.clean_command()

        # save the watchpoint_hits data
        watchpoint_hits = []
        watchpoint_stream = self._cache_store.get_stream_handler(Streams.WATCHPOINT)
        graph_stream = self._cache_store.get_stream_handler(Streams.GRAPH).get_graph_handler_by_rank_id(0)
        for watchpoint_hit_proto in request_iterator:
            node_full_name = watchpoint_hit_proto.tensor.node_name
            graph_name = graph_stream.get_graph_id_by_full_name(node_full_name)
            if not graph_name:
                log.warning("Cannot find node %s in graph. Skip it.", node_full_name)
                continue
            ui_node_name = graph_stream.get_node_name_by_full_name(node_full_name, graph_name)
            log.debug("Receive watch point hit: %s", watchpoint_hit_proto)
            if not ui_node_name:
                log.info("Not support to show %s on graph.", node_full_name)
                continue
            watchpoint_hit = {
                'tensor_proto': watchpoint_hit_proto.tensor,
                'watchpoint': copy.deepcopy(watchpoint_stream.get_watchpoint_by_id(watchpoint_hit_proto.id)),
                'node_name': ui_node_name,
                'graph_name': graph_name
            }
            hit_params = {}
            for param in watchpoint_hit_proto.watch_condition.params:
                if param.name not in (ParamNameEnum.RTOL.value, ParamNameEnum.RANGE_START_INCLUSIVE.value,
                                      ParamNameEnum.RANGE_END_INCLUSIVE.value) \
                        and watchpoint_hit_proto.error_code == 0:
                    hit_params[param.name] = param.actual_value
            watchpoint = watchpoint_hit.get('watchpoint', None)
            if not watchpoint:
                return get_ack_reply()
            params = watchpoint.condition.get('params', [])
            for i, param in enumerate(params):
                name = param['name']
                if name in hit_params.keys():
                    watchpoint_hit['watchpoint'].condition['params'][i]['actual_value'] = hit_params[name]
                else:
                    watchpoint_hit['watchpoint'].condition['params'][i]['actual_value'] = None
            watchpoint_hit['error_code'] = watchpoint_hit_proto.error_code
            watchpoint_hits.append(watchpoint_hit)
        self._received_hit = watchpoint_hits
        reply = get_ack_reply()
        return reply

    def SendHeartbeat(self, request, context):
        """Deal with heartbeat sent from training client."""
        # only support single card training now
        if self._heartbeat is None or not self._heartbeat.is_alive():
            period = request.period
            min_period_seconds = 5
            max_period_seconds = 3600
            if not min_period_seconds <= period <= max_period_seconds:
                log.error("Invalid period time which should be in [5, 3600].")
                return get_ack_reply(-1)
            self._heartbeat = HeartbeatListener(self._cache_store, request.period)
            self._heartbeat.start()
        self._heartbeat.send()
        return get_ack_reply()


class HeartbeatListener(threading.Thread):
    """Heartbeat listener."""
    _MAX_TRY_EXCEPT_COUNT = 3

    def __init__(self, cache_store, period=None):
        """
        Initialize Heartbeat listener object.

        Args:
            cache_store (DebuggerCacheStore): Cache store for debugger server.
            period (int): The max waiting seconds for each period.
        """
        super(HeartbeatListener, self).__init__()
        self._heartbeat_queue = Queue(maxsize=1)
        self._cache_store = cache_store
        # the waiting period time for next heartbeat, default waiting for 15 seconds
        time_delay = 0.5
        self._period = period + time_delay if period else 15
        self._running = threading.Event()
        self._running.clear()

    @debugger_server_wrap
    def run(self):
        """Function that should be called when thread started."""
        self._running.set()
        log.info("Start listening for heartbeat.")
        try_count = 0
        while self._running.is_set():
            try:
                self._heartbeat_queue.get(timeout=self._period)
                # reset try count if received heartbeat
                try_count = 0
            except Empty:
                try_count += 1
                if try_count >= self._MAX_TRY_EXCEPT_COUNT:
                    break
                log.info("Missing heartbeat. Try again.")

        log.warning("Missing heartbeat. Reset online session.")
        self._cache_store.clean()
        metadata_stream = self._cache_store.get_stream_handler(Streams.METADATA)
        # put metadata into data queue
        metadata = metadata_stream.get()
        self._cache_store.put_data(metadata)

    def send(self):
        """Send heartbeat signal."""
        try:
            self._heartbeat_queue.put(True, block=False)
        except Full:
            log.debug("Too many heartbeat received.")

    @debugger_server_wrap
    def stop(self):
        """Stop debugger server."""
        self._running.wait()
        log.info("Start to stop offline debugger server.")
        self._running.clear()
        self.send()
        self.join()


def _convert_tensor_base(tensor_data):
    """Convert tensor object to dict."""
    tensor_data_dict = {
        'dtype': tensor_data.data_type,
        'shape': tensor_data.shape,
        'data_size': tensor_data.data_size
    }
    return tensor_data_dict
